import base64
import os
import unittest
import logging
from multiprocessing import get_context
from concurrent.futures.thread import ThreadPoolExecutor
from typing import List, Dict, Tuple, Any, Union

from tonclient.bindings.types import TCResponseType
from tonclient.test.helpers import SAMPLES_DIR, async_custom_client, send_grams
from tonclient.types import Abi, Signer, CallSet, DeploySet, DebotAction, \
    DebotState, AppRequestResult, ParamsOfAppDebotBrowser, \
    ResultOfAppDebotBrowser, ParamsOfEncodeMessage, ParamsOfProcessMessage, \
    ParamsOfStart, ParamsOfFetch, ParamsOfExecute, ParamsOfAppRequest, \
    ParamsOfResolveAppRequest, KeyPair


class TestTonDebotAsyncCore(unittest.TestCase):
    def setUp(self) -> None:
        self.keypair = async_custom_client.crypto.generate_random_sign_keys()
        self.target_abi = Abi.from_path(
            path=os.path.join(SAMPLES_DIR, 'DebotTarget.abi.json'))
        self.debot_abi = Abi.from_path(
            path=os.path.join(SAMPLES_DIR, 'Debot.abi.json'))
        with open(os.path.join(SAMPLES_DIR, 'DebotTarget.tvc'), 'rb') as fp:
            self.target_tvc = base64.b64encode(fp.read()).decode()
        with open(os.path.join(SAMPLES_DIR, 'Debot.tvc'), 'rb') as fp:
            self.debot_tvc = base64.b64encode(fp.read()).decode()

        self.debot_address, self.target_address = self.__deploy()

    def test_goto(self):
        steps = [
            {'choice': 0, 'inputs': [], 'outputs': ['Test Goto Action']},
            {'choice': 0, 'inputs': [], 'outputs': ['Debot Tests']},
            {'choice': 7, 'inputs': [], 'outputs': []}
        ]
        params = ParamsOfStart(address=self.debot_address)
        debot_browser(
            steps=steps, params=params, start_fn='start', keypair=self.keypair)

    def test_print(self):
        steps = [
            {'choice': 1, 'inputs': [], 'outputs': ['Test Print Action', 'test2: instant print', 'test instant print']},
            {'choice': 0, 'inputs': [], 'outputs': ['test simple print']},
            {'choice': 1, 'inputs': [], 'outputs': [f'integer=1,addr={self.target_address},string=test_string_1']},
            {'choice': 2, 'inputs': [], 'outputs': ['Debot Tests']},
            {'choice': 7, 'inputs': [], 'outputs': []}
        ]
        params = ParamsOfStart(address=self.debot_address)
        debot_browser(
            steps=steps, params=params, start_fn='start', keypair=self.keypair)

    def test_run_action(self):
        steps = [
            {'choice': 2, 'inputs': [], 'outputs': ['Test Run Action']},
            {'choice': 0, 'inputs': ['-1:1111111111111111111111111111111111111111111111111111111111111111'], 'outputs': ['Test Instant Run', 'test1: instant run 1', 'test2: instant run 2']},
            {'choice': 0, 'inputs': [], 'outputs': ['Test Run Action']},
            {'choice': 1, 'inputs': ['hello'], 'outputs': []},
            {'choice': 2, 'inputs': [], 'outputs': ['integer=2,addr=-1:1111111111111111111111111111111111111111111111111111111111111111,string=hello']},
            {'choice': 3, 'inputs': [], 'outputs': ['Debot Tests']},
            {'choice': 7, 'inputs': [], 'outputs': []}
        ]
        params = ParamsOfStart(address=self.debot_address)
        debot_browser(
            steps=steps, params=params, start_fn='start', keypair=self.keypair)

    def test_run_method(self):
        steps = [
            {'choice': 3, 'inputs': [], 'outputs': ['Test Run Method Action']},
            {'choice': 0, 'inputs': [], 'outputs': []},
            {'choice': 1, 'inputs': [], 'outputs': ['data=64']},
            {'choice': 2, 'inputs': [], 'outputs': ['Debot Tests']},
            {'choice': 7, 'inputs': [], 'outputs': []}
        ]
        params = ParamsOfStart(address=self.debot_address)
        debot_browser(
            steps=steps, params=params, start_fn='start', keypair=self.keypair)

    def test_send_message(self):
        steps = [
            {'choice': 4, 'inputs': [], 'outputs': ['Test Send Msg Action']},
            {'choice': 0, 'inputs': [], 'outputs': ['Sending message {}', 'Transaction succeeded.']},
            {'choice': 1, 'inputs': [], 'outputs': []},
            {'choice': 2, 'inputs': [], 'outputs': ['data=100']},
            {'choice': 3, 'inputs': [], 'outputs': ['Debot Tests']},
            {'choice': 7, 'inputs': [], 'outputs': []}
        ]
        params = ParamsOfStart(address=self.debot_address)
        debot_browser(
            steps=steps, params=params, start_fn='start', keypair=self.keypair)

    def test_invoke(self):
        steps = [
            {'choice': 5, 'inputs': [self.debot_address], 'outputs': ['Test Invoke Debot Action', 'enter debot address:']},
            {'choice': 0, 'inputs': [self.debot_address], 'outputs': ['Test Invoke Debot Action', 'enter debot address:'], 'invokes': [
                {'choice': 0, 'inputs': [], 'outputs': ['Print test string', 'Debot is invoked']},
                {'choice': 0, 'inputs': [], 'outputs': ['Sending message {}', 'Transaction succeeded.']}
            ]},
            {'choice': 1, 'inputs': [], 'outputs': ['Debot Tests']},
            {'choice': 7, 'inputs': [], 'outputs': []}
        ]
        params = ParamsOfStart(address=self.debot_address)
        debot_browser(
            steps=steps, params=params, start_fn='start', keypair=self.keypair)

    def test_engine_calls(self):
        steps = [
            {'choice': 6, 'inputs': [], 'outputs': ['Test Engine Calls']},
            {'choice': 0, 'inputs': [], 'outputs': []},
            {'choice': 1, 'inputs': [], 'outputs': []},
            # {'choice': 2, 'inputs': [], 'outputs': []},
            {'choice': 3, 'inputs': [], 'outputs': []},
            {'choice': 4, 'inputs': [], 'outputs': []},
            {'choice': 5, 'inputs': [], 'outputs': ['Debot Tests']},
            {'choice': 7, 'inputs': [], 'outputs': []}
        ]
        params = ParamsOfStart(address=self.debot_address)
        debot_browser(
            steps=steps, params=params, start_fn='start', keypair=self.keypair)

    @staticmethod
    def debot_print_state(state: Dict[str, Any]):
        # Print messages
        for message in state['messages']:
            logging.info(f'[LOG]\t{message}')
        # Print available actions
        for action in state['actions']:
            logging.info(f'[ACTION]\t{action}')

    def __deploy(self) -> Tuple[str, str]:
        """ Deploy debot and target """
        signer = Signer.Keys(keys=self.keypair)

        # Deploy target
        call_set = CallSet(function_name='constructor')
        deploy_set = DeploySet(tvc=self.target_tvc)
        encode_params = ParamsOfEncodeMessage(
            abi=self.target_abi, signer=signer, deploy_set=deploy_set,
            call_set=call_set)
        message = async_custom_client.abi.encode_message(params=encode_params)
        send_grams(address=message.address)
        process_params = ParamsOfProcessMessage(
            message_encode_params=encode_params, send_events=False)
        target = async_custom_client.processing.process_message(
            params=process_params)

        # Deploy debot
        call_set = CallSet(
            function_name='constructor', input={
                'debotAbi': self.debot_abi.value.encode().hex(),
                'targetAbi': self.target_abi.value.encode().hex(),
                'targetAddr': target.transaction['account_addr']
            })
        deploy_set = DeploySet(tvc=self.debot_tvc)
        encode_params = ParamsOfEncodeMessage(
            abi=self.debot_abi, signer=signer, deploy_set=deploy_set,
            call_set=call_set)
        message = async_custom_client.abi.encode_message(params=encode_params)
        send_grams(address=message.address)
        process_params = ParamsOfProcessMessage(
            message_encode_params=encode_params, send_events=False)
        debot = async_custom_client.processing.process_message(
            params=process_params)

        return (
            debot.transaction['account_addr'],
            target.transaction['account_addr']
        )


def debot_browser(
        steps: List[Dict[str, Any]], start_fn: str, keypair: KeyPair,
        params: Union[ParamsOfStart, ParamsOfFetch],
        actions: List[DebotAction] = None):

    def __callback(response_data, response_type, *args):
        # Process notifications
        if response_type == TCResponseType.AppNotify:
            notify = ParamsOfAppDebotBrowser.from_dict(data=response_data)

            # Process notification types
            if isinstance(notify, ParamsOfAppDebotBrowser.Log):
                state['messages'].append(notify.msg)
            if isinstance(notify, ParamsOfAppDebotBrowser.Switch):
                state['switch_started'] = True
                if notify.context_id == DebotState.EXIT:
                    state['finished'] = True
                state['actions'].clear()
            if isinstance(
                    notify, ParamsOfAppDebotBrowser.SwitchCompleted):
                state['switch_started'] = False
            if isinstance(notify, ParamsOfAppDebotBrowser.ShowAction):
                state['actions'].append(notify.action)

        # Process requests
        if response_type == TCResponseType.AppRequest:
            request = ParamsOfAppRequest(**response_data)
            request_data = ParamsOfAppDebotBrowser.from_dict(
                data=request.request_data)
            result = None

            # Process request types
            if isinstance(request_data, ParamsOfAppDebotBrowser.Input):
                result = ResultOfAppDebotBrowser.Input(
                    value=state['step']['inputs'][0])
            if isinstance(
                    request_data, ParamsOfAppDebotBrowser.GetSigningBox):
                with ThreadPoolExecutor() as executor:
                    future = executor.submit(
                        async_custom_client.crypto.get_signing_box,
                        params=keypair)
                    signing_box = future.result()
                result = ResultOfAppDebotBrowser.GetSigningBox(
                    signing_box=signing_box.handle)
            if isinstance(
                    request_data, ParamsOfAppDebotBrowser.InvokeDebot):
                invoke_steps = state['step']['invokes']
                fetch_params = ParamsOfFetch(address=request_data.debot_addr)

                # Here we should call `debot_browser` in `spawn` mode
                # subprocess for compatibility with `Unix` systems.
                # MacOS, Windows use `spawn` by default
                with get_context('spawn').Pool() as pool:
                    pool.apply(debot_browser, kwds={
                        'steps': invoke_steps,
                        'params': fetch_params,
                        'start_fn': 'fetch',
                        'actions': [request_data.action],
                        'keypair': keypair
                    })
                result = ResultOfAppDebotBrowser.InvokeDebot()

            # Resolve app request
            result = AppRequestResult.Ok(result=result.dict)
            resolve_params = ParamsOfResolveAppRequest(
                app_request_id=request.app_request_id, result=result)
            with ThreadPoolExecutor() as executor:
                future = executor.submit(
                    async_custom_client.resolve_app_request,
                    params=resolve_params)
                future.result()

    # Create initial state
    state: Dict[str, Any] = {
        'messages': [],
        'actions': actions or [],
        'steps': steps,
        'step': None,
        'switch_started': False,
        'finished': False
    }

    # Start debot browser and get handle
    debot = getattr(async_custom_client.debot, start_fn)(
        params=params, callback=__callback)
    TestTonDebotAsyncCore.debot_print_state(state=state)

    while not state['finished']:
        step: Dict[str, Any] = state['steps'].pop(0)
        action = state['actions'][step['choice']]
        logging.info(f'[ACTION SELECTED]\t{action}')
        state['messages'].clear()
        state['step'] = step

        exec_params = ParamsOfExecute(
            debot_handle=debot.debot_handle, action=action)
        async_custom_client.debot.execute(params=exec_params)
        TestTonDebotAsyncCore.debot_print_state(state=state)

        test_case = unittest.TestCase()
        test_case.assertEqual(
            len(state['step']['outputs']), len(state['messages']))

        if not len(state['steps']):
            break

    async_custom_client.debot.remove(params=debot)



import base64

import unittest

from tonclient.errors import TonException
from tonclient.test.helpers import async_core_client, sync_core_client
from tonclient.types import KeyPair, MnemonicDictionary, ParamsOfHash, \
    ParamsOfHDKeyXPrvFromMnemonic, ParamsOfHDKeySecretFromXPrv, \
    ParamsOfHDKeyPublicFromXPrv, ParamsOfHDKeyDeriveFromXPrv, \
    ParamsOfHDKeyDeriveFromXPrvPath, ParamsOfConvertPublicKeyToTonSafeFormat, \
    ParamsOfSign, ParamsOfVerifySignature, ParamsOfModularPower, \
    ParamsOfFactorize, ParamsOfTonCrc16, ParamsOfGenerateRandomBytes, \
    ParamsOfMnemonicWords, ParamsOfMnemonicFromRandom, \
    ParamsOfMnemonicFromEntropy, ParamsOfMnemonicVerify, \
    ParamsOfMnemonicDeriveSignKeys, ParamsOfNaclSignKeyPairFromSecret, \
    ParamsOfNaclSign, ParamsOfNaclSignOpen, ParamsOfNaclBoxKeyPairFromSecret, \
    ParamsOfNaclBox, ParamsOfNaclBoxOpen, ParamsOfNaclSecretBox, \
    ParamsOfNaclSecretBoxOpen, ParamsOfScrypt, ParamsOfChaCha20, \
    ParamsOfSigningBoxSign, ParamsOfAppRequest, ParamsOfAppSigningBox, \
    ParamsOfResolveAppRequest, ResultOfAppSigningBox, AppRequestResult


class TestTonCryptoAsyncCore(unittest.TestCase):
    def setUp(self) -> None:
        self.mnemonic = 'abuse boss fly battle rubber wasp afraid hamster guide essence vibrant tattoo'
        self.master_xprv = 'xprv9s21ZrQH143K25JhKqEwvJW7QAiVvkmi4WRenBZanA6kxHKtKAQQKwZG65kCyW5jWJ8NY9e3GkRoistUjjcpHNsGBUv94istDPXvqGNuWpC'

    def test_sha256(self):
        params = ParamsOfHash(
            data=base64.b64encode('TON is our future'.encode()).decode())
        result = async_core_client.crypto.sha256(params=params)
        self.assertEqual(
            '1e7fd5ec201652b5375e5edf3e86d0513394d2c2004dd506415abf0578261951',
            result.hash)

        params.data = base64.b64encode(
            bytes.fromhex('544f4e206973206f757220667574757265')).decode()
        result = async_core_client.crypto.sha256(params=params)
        self.assertEqual(
            '1e7fd5ec201652b5375e5edf3e86d0513394d2c2004dd506415abf0578261951',
            result.hash)

        params.data = 'VE9OIGlzIG91ciBmdXR1cmU='
        result = async_core_client.crypto.sha256(params=params)
        self.assertEqual(
            '1e7fd5ec201652b5375e5edf3e86d0513394d2c2004dd506415abf0578261951',
            result.hash)

    def test_sha512(self):
        data = base64.b64encode('TON is our future'.encode()).decode()
        params = ParamsOfHash(data=data)
        result = async_core_client.crypto.sha512(params=params)
        self.assertEqual(
            '4c52dd4cefc68319bac5e97c1f0d18ae8194fb0dd8d9e090ba8376834a0756175a9a736d1e69cb1a58d25c3d554b02a2b8ed9c3ae5cbeeccc3277746a363a434',
            result.hash)

    def test_hdkey_xprv_from_mnemonic(self):
        params = ParamsOfHDKeyXPrvFromMnemonic(phrase=self.mnemonic)
        result = async_core_client.crypto.hdkey_xprv_from_mnemonic(
            params=params)
        self.assertEqual(self.master_xprv, result.xprv)

        with self.assertRaises(TonException):
            params.phrase = 0
            async_core_client.crypto.hdkey_xprv_from_mnemonic(params=params)

    def test_hdkey_secret_from_xprv(self):
        params = ParamsOfHDKeySecretFromXPrv(xprv=self.master_xprv)
        result = async_core_client.crypto.hdkey_secret_from_xprv(params=params)
        self.assertEqual(
            '0c91e53128fa4d67589d63a6c44049c1068ec28a63069a55ca3de30c57f8b365',
            result.secret)

        with self.assertRaises(TonException):
            params.xprv = ''
            async_core_client.crypto.hdkey_secret_from_xprv(params=params)

    def test_hdkey_public_from_xprv(self):
        params = ParamsOfHDKeyPublicFromXPrv(xprv=self.master_xprv)
        result = async_core_client.crypto.hdkey_public_from_xprv(params=params)
        self.assertEqual(
            '02a8eb63085f73c33fa31b4d1134259406347284f8dab6fc68f4bf8c96f6c39b75',
            result.public)

    def test_hdkey_derive_from_xprv(self):
        params = ParamsOfHDKeyDeriveFromXPrv(
            xprv=self.master_xprv, child_index=0, hardened=False)
        result = async_core_client.crypto.hdkey_derive_from_xprv(params=params)
        self.assertEqual(
            'xprv9uZwtSeoKf1swgAkVVCEUmC2at6t7MCJoHnBbn1MWJZyxQ4cySkVXPyNh7zjf9VjsP4vEHDDD2a6R35cHubg4WpzXRzniYiy8aJh1gNnBKv',
            result.xprv)

        with self.assertRaises(TonException):
            params.child_index = -1
            async_core_client.crypto.hdkey_derive_from_xprv(params=params)

    def test_hdkey_derive_from_xprv_path(self):
        params = ParamsOfHDKeyDeriveFromXPrvPath(
            xprv=self.master_xprv, path="m/44'/60'/0'/0'")
        result = async_core_client.crypto.hdkey_derive_from_xprv_path(
            params=params)
        self.assertEqual(
            'xprvA1KNMo63UcGjmDF1bX39Cw2BXGUwrwMjeD5qvQ3tA3qS3mZQkGtpf4DHq8FDLKAvAjXsYGLHDP2dVzLu9ycta8PXLuSYib2T3vzLf3brVgZ',
            result.xprv)

        with self.assertRaises(TonException):
            params.path = 'm/'
            async_core_client.crypto.hdkey_derive_from_xprv_path(params=params)

    def test_convert_public_key_to_ton_safe_format(self):
        params = ParamsOfConvertPublicKeyToTonSafeFormat(
            public_key='06117f59ade83e097e0fb33e5d29e8735bda82b3bf78a015542aaa853bb69600')
        safe = async_core_client.crypto.convert_public_key_to_ton_safe_format(
            params=params)
        self.assertEqual(
            'PuYGEX9Zreg-CX4Psz5dKehzW9qCs794oBVUKqqFO7aWAOTD',
            safe.ton_public_key)

        with self.assertRaises(TonException):
            params.public_key = None
            async_core_client.crypto.convert_public_key_to_ton_safe_format(
                params=params)

    def test_generate_random_sign_keys(self):
        keypair = async_core_client.crypto.generate_random_sign_keys()
        self.assertEqual(64, len(keypair.public))
        self.assertEqual(64, len(keypair.secret))
        self.assertNotEqual(keypair.secret, keypair.public)

    def test_sign_and_verify(self):
        unsigned = base64.b64encode('Test Message'.encode()).decode()
        keypair = KeyPair(
            public='1869b7ef29d58026217e9cf163cbfbd0de889bdf1bf4daebf5433a312f5b8d6e',
            secret='56b6a77093d6fdf14e593f36275d872d75de5b341942376b2a08759f3cbae78f')

        # Sign message
        sign_params = ParamsOfSign(unsigned=unsigned, keys=keypair)
        signed = async_core_client.crypto.sign(params=sign_params)
        self.assertEqual(
            '+wz+QO6l1slgZS5s65BNqKcu4vz24FCJz4NSAxef9lu0jFfs8x3PzSZRC+pn5k8+aJi3xYMA3BQzglQmjK3hA1Rlc3QgTWVzc2FnZQ==',
            signed.signed)
        self.assertEqual(
            'fb0cfe40eea5d6c960652e6ceb904da8a72ee2fcf6e05089cf835203179ff65bb48c57ecf31dcfcd26510bea67e64f3e6898b7c58300dc14338254268cade103',
            signed.signature)

        # Verify signature
        verify_params = ParamsOfVerifySignature(
            signed=signed.signed, public=keypair.public)
        verified = async_core_client.crypto.verify_signature(
            params=verify_params)
        self.assertEqual(unsigned, verified.unsigned)
        self.assertEqual(
            base64.b64decode(unsigned.encode()),
            base64.b64decode(verified.unsigned.encode()))

        with self.assertRaises(TonException):
            sign_params.keys = KeyPair(public='1', secret='2')
            async_core_client.crypto.sign(params=sign_params)

        with self.assertRaises(TonException):
            verify_params.signed = 'simple'
            async_core_client.crypto.verify_signature(params=verify_params)

    def test_modular_power(self):
        params = ParamsOfModularPower(
            base='0123456789ABCDEF', exponent='0123', modulus='01234567')
        result = async_core_client.crypto.modular_power(params=params)
        self.assertEqual('63bfdf', result.modular_power)

        with self.assertRaises(TonException):
            params.base = '1'
            params.modulus = '0.2'
            async_core_client.crypto.modular_power(params=params)

    def test_factorize(self):
        params = ParamsOfFactorize(composite='17ED48941A08F981')
        result = async_core_client.crypto.factorize(params=params)
        self.assertIsInstance(result.factors, list)
        self.assertEqual('494C553B', result.factors[0])
        self.assertEqual('53911073', result.factors[1])

        with self.assertRaises(TonException):
            params.composite = 'a3'
            async_core_client.crypto.factorize(params=params)

    def test_ton_crc16(self):
        params = ParamsOfTonCrc16(
            data=base64.b64encode(bytes.fromhex('0123456789abcdef')).decode())
        result = async_core_client.crypto.ton_crc16(params=params)
        self.assertEqual(43349, result.crc)

        with self.assertRaises(TonException):
            params.data = '0=='
            async_core_client.crypto.ton_crc16(params=params)

    def test_generate_random_bytes(self):
        params = ParamsOfGenerateRandomBytes(length=32)
        result = async_core_client.crypto.generate_random_bytes(params=params)
        self.assertEqual(44, len(result.bytes))
        bts = base64.b64decode(result.bytes.encode())
        self.assertEqual(32, len(bts))

        with self.assertRaises(TonException):
            params.length = '1'
            async_core_client.crypto.generate_random_bytes(params=params)

    def test_mnemonic_words(self):
        params = ParamsOfMnemonicWords()
        result = async_core_client.crypto.mnemonic_words(params=params)
        self.assertEqual(2048, len(result.words.split(' ')))

        with self.assertRaises(TonException):
            params.dictionary = 100
            async_core_client.crypto.mnemonic_words(params=params)

    def test_mnemonic_from_random(self):
        params = ParamsOfMnemonicFromRandom()
        result = async_core_client.crypto.mnemonic_from_random(params=params)
        self.assertEqual(12, len(result.phrase.split(' ')))

        for d in range(1, 8):
            for count in [12, 15, 18, 21, 24]:
                params.dictionary = list(MnemonicDictionary)[d]
                params.word_count = count
                result = async_core_client.crypto.mnemonic_from_random(
                    params=params)
                self.assertEqual(count, len(result.phrase.split(' ')))

        with self.assertRaises(TonException):
            params.word_count = 0
            async_core_client.crypto.mnemonic_from_random(params=params)

    def test_mnemonic_from_entropy(self):
        params = ParamsOfMnemonicFromEntropy(
            entropy='00112233445566778899AABBCCDDEEFF')
        result = async_core_client.crypto.mnemonic_from_entropy(params=params)
        self.assertEqual(
            'abandon math mimic master filter design carbon crystal rookie group knife young',
            result.phrase)

        with self.assertRaises(TonException):
            params.entropy = '01'
            async_core_client.crypto.mnemonic_from_entropy(params=params)

    def test_mnemonic_verify(self):
        m_params = ParamsOfMnemonicFromRandom()
        result = async_core_client.crypto.mnemonic_from_random(params=m_params)

        v_params = ParamsOfMnemonicVerify(phrase=result.phrase)
        result = async_core_client.crypto.mnemonic_verify(params=v_params)
        self.assertEqual(True, result.valid)

        for d in range(1, 8):
            for count in [12, 15, 18, 21, 24]:
                m_params.dictionary = list(MnemonicDictionary)[d]
                m_params.word_count = count
                mnemonic = async_core_client.crypto.mnemonic_from_random(
                    params=m_params)

                v_params.phrase = mnemonic.phrase
                v_params.dictionary = m_params.dictionary
                v_params.word_count = m_params.word_count
                result = async_core_client.crypto.mnemonic_verify(
                    params=v_params)
                self.assertEqual(True, result.valid)

        v_params = ParamsOfMnemonicVerify(phrase='one')
        result = async_core_client.crypto.mnemonic_verify(params=v_params)
        self.assertEqual(False, result.valid)

    def test_mnemonic_derive_sign_keys(self):
        # Derive from random phrase
        params = ParamsOfMnemonicFromRandom()
        mnemonic = async_core_client.crypto.mnemonic_from_random(params=params)

        params = ParamsOfMnemonicDeriveSignKeys(phrase=mnemonic.phrase)
        keypair = async_core_client.crypto.mnemonic_derive_sign_keys(
            params=params)
        self.assertIsInstance(keypair, KeyPair)

        # Derive from provided phrase and convert public to ton_safe
        phrase = 'unit follow zone decline glare flower crisp vocal adapt magic much mesh cherry teach mechanic rain float vicious solution assume hedgehog rail sort chuckle'
        derive_params = ParamsOfMnemonicDeriveSignKeys(
            phrase=phrase, dictionary=MnemonicDictionary.TON, word_count=24)
        keypair = async_core_client.crypto.mnemonic_derive_sign_keys(
            params=derive_params)

        convert_params = ParamsOfConvertPublicKeyToTonSafeFormat(
            public_key=keypair.public)
        result = async_core_client.crypto.convert_public_key_to_ton_safe_format(
            params=convert_params)
        self.assertEqual(
            'PuYTvCuf__YXhp-4jv3TXTHL0iK65ImwxG0RGrYc1sP3H4KS',
            result.ton_public_key)

        # Derive with path
        derive_params = ParamsOfMnemonicDeriveSignKeys(
            phrase=phrase, path='m', dictionary=MnemonicDictionary.TON,
            word_count=24)
        keypair = async_core_client.crypto.mnemonic_derive_sign_keys(
            params=derive_params)

        convert_params = ParamsOfConvertPublicKeyToTonSafeFormat(
            public_key=keypair.public)
        result = async_core_client.crypto.convert_public_key_to_ton_safe_format(
            params=convert_params)
        self.assertEqual(
            'PubDdJkMyss2qHywFuVP1vzww0TpsLxnRNnbifTCcu-XEgW0',
            result.ton_public_key)

        # Derive from 12-word phrase
        phrase = 'abandon math mimic master filter design carbon crystal rookie group knife young'
        derive_params = ParamsOfMnemonicDeriveSignKeys(phrase=phrase)
        keypair = async_core_client.crypto.mnemonic_derive_sign_keys(
            params=derive_params)

        convert_params = ParamsOfConvertPublicKeyToTonSafeFormat(
            public_key=keypair.public)
        result = async_core_client.crypto.convert_public_key_to_ton_safe_format(
            params=convert_params)
        self.assertEqual(
            'PuZhw8W5ejPJwKA68RL7sn4_RNmeH4BIU_mEK7em5d4_-cIx',
            result.ton_public_key)

        # Derive from mnemonic from entropy
        params = ParamsOfMnemonicFromEntropy(
            entropy='2199ebe996f14d9e4e2595113ad1e627')
        mnemonic = async_core_client.crypto.mnemonic_from_entropy(
            params=params)

        derive_params = ParamsOfMnemonicDeriveSignKeys(phrase=mnemonic.phrase)
        keypair = async_core_client.crypto.mnemonic_derive_sign_keys(
            params=derive_params)

        convert_params = ParamsOfConvertPublicKeyToTonSafeFormat(
            public_key=keypair.public)
        result = async_core_client.crypto.convert_public_key_to_ton_safe_format(
            params=convert_params)
        self.assertEqual(
            'PuZdw_KyXIzo8IksTrERN3_WoAoYTyK7OvM-yaLk711sUIB3',
            result.ton_public_key)

    def test_nacl_sign_keypair_from_secret_key(self):
        params = ParamsOfNaclSignKeyPairFromSecret(
            secret='8fb4f2d256e57138fb310b0a6dac5bbc4bee09eb4821223a720e5b8e1f3dd674')
        keypair = async_core_client.crypto.nacl_sign_keypair_from_secret_key(
            params=params)
        self.assertEqual(
            'aa5533618573860a7e1bf19f34bd292871710ed5b2eafa0dcdbb33405f2231c6',
            keypair.public)

        with self.assertRaises(TonException):
            params.secret = '0a'
            async_core_client.crypto.nacl_sign_keypair_from_secret_key(
                params=params)

    def test_nacl_sign(self):
        # Nacl sign data
        unsigned = base64.b64encode('Test Message'.encode()).decode()
        secret = '56b6a77093d6fdf14e593f36275d872d75de5b341942376b2a08759f3cbae78f1869b7ef29d58026217e9cf163cbfbd0de889bdf1bf4daebf5433a312f5b8d6e'

        params = ParamsOfNaclSign(unsigned=unsigned, secret=secret)
        signed = async_core_client.crypto.nacl_sign(params=params)
        self.assertEqual(
            '+wz+QO6l1slgZS5s65BNqKcu4vz24FCJz4NSAxef9lu0jFfs8x3PzSZRC+pn5k8+aJi3xYMA3BQzglQmjK3hA1Rlc3QgTWVzc2FnZQ==',
            signed.signed)

        # Nacl sign open
        params = ParamsOfNaclSignOpen(
            signed=signed.signed,
            public='1869b7ef29d58026217e9cf163cbfbd0de889bdf1bf4daebf5433a312f5b8d6e')
        result = async_core_client.crypto.nacl_sign_open(params=params)
        self.assertEqual(unsigned, result.unsigned)

        # Nacl sign detached
        params = ParamsOfNaclSign(unsigned=unsigned, secret=secret)
        result = async_core_client.crypto.nacl_sign_detached(params=params)
        self.assertEqual(
            'fb0cfe40eea5d6c960652e6ceb904da8a72ee2fcf6e05089cf835203179ff65bb48c57ecf31dcfcd26510bea67e64f3e6898b7c58300dc14338254268cade103',
            result.signature)

        with self.assertRaises(TonException):
            params.secret = '0=='
            async_core_client.crypto.nacl_sign(params=params)

        with self.assertRaises(TonException):
            params = ParamsOfNaclSignOpen(signed=signed.signed, public='0x00')
            async_core_client.crypto.nacl_sign_open(params=params)

        with self.assertRaises(TonException):
            params = ParamsOfNaclSign(unsigned='100', secret=secret)
            params.unsigned = 100
            async_core_client.crypto.nacl_sign_detached(params=params)

    def test_nacl_box_keypair(self):
        keypair = async_core_client.crypto.nacl_box_keypair()
        self.assertEqual(64, len(keypair.public))
        self.assertEqual(64, len(keypair.secret))
        self.assertNotEqual(keypair.public, keypair.secret)

    def test_nacl_box_keypair_from_secret_key(self):
        params = ParamsOfNaclBoxKeyPairFromSecret(
            secret='e207b5966fb2c5be1b71ed94ea813202706ab84253bdf4dc55232f82a1caf0d4')
        keypair = async_core_client.crypto.nacl_box_keypair_from_secret_key(
            params=params)
        self.assertEqual(
            'a53b003d3ffc1e159355cb37332d67fc235a7feb6381e36c803274074dc3933a',
            keypair.public)

        with self.assertRaises(TonException):
            params.secret = '0x00'
            async_core_client.crypto.nacl_box_keypair_from_secret_key(
                params=params)

    def test_nacl_box_and_open(self):
        decrypted = base64.b64encode('Test Message'.encode()).decode()
        nonce = 'cd7f99924bf422544046e83595dd5803f17536f5c9a11746'
        their_public = 'c4e2d9fe6a6baf8d1812b799856ef2a306291be7a7024837ad33a8530db79c6b'
        secret = 'd9b9dc5033fb416134e5d2107fdbacab5aadb297cb82dbdcd137d663bac59f7f'

        # Create nacl box
        box_params = ParamsOfNaclBox(
            decrypted=decrypted, nonce=nonce, their_public=their_public,
            secret=secret)
        box = async_core_client.crypto.nacl_box(params=box_params)
        self.assertEqual(
            'li4XED4kx/pjQ2qdP0eR2d/K30uN94voNADxwA==', box.encrypted)

        # Open nacl box
        box_open_params = ParamsOfNaclBoxOpen(
            encrypted=box.encrypted, nonce=nonce, their_public=their_public,
            secret=secret)
        opened = async_core_client.crypto.nacl_box_open(params=box_open_params)
        self.assertEqual(decrypted, opened.decrypted)

        with self.assertRaises(TonException):
            box_params.decrypted = '0x00'
            box_params.their_public = ''
            async_core_client.crypto.nacl_box(params=box_params)

        with self.assertRaises(TonException):
            box_open_params.secret = ''
            async_core_client.crypto.nacl_box_open(params=box_open_params)

    def test_nacl_secret_box_and_open(self):
        decrypted = base64.b64encode(
            'Test Message \' \" {} $=,?'.encode()).decode()
        nonce = '2a33564717595ebe53d91a785b9e068aba625c8453a76e45'
        key = '8f68445b4e78c000fe4d6b7fc826879c1e63e3118379219a754ae66327764bd8'

        # Create nacl secret box
        box_params = ParamsOfNaclSecretBox(
            decrypted=decrypted, nonce=nonce, key=key)
        box = async_core_client.crypto.nacl_secret_box(params=box_params)
        self.assertEqual(
            'I6QZteixTdul0K0ldT+/U4QF0t/C1Q8RGyzQ2Hl7886DpW3/DK5ijg==',
            box.encrypted)

        # Open nacl secret box
        box_open_params = ParamsOfNaclSecretBoxOpen(
            encrypted=box.encrypted, nonce=nonce, key=key)
        opened = async_core_client.crypto.nacl_secret_box_open(
            params=box_open_params)
        self.assertEqual(decrypted, opened.decrypted)

        with self.assertRaises(TonException):
            box_params.decrypted = '0x00'
            box_params.key = None
            async_core_client.crypto.nacl_secret_box(params=box_params)

        with self.assertRaises(TonException):
            box_open_params.key = ''
            async_core_client.crypto.nacl_secret_box_open(
                params=box_open_params)

    def test_scrypt(self):
        password = base64.b64encode('Test Password'.encode()).decode()
        salt = base64.b64encode('Test Salt'.encode()).decode()

        params = ParamsOfScrypt(
            password=password, salt=salt, log_n=10, r=8, p=16, dk_len=64)
        result = async_core_client.crypto.scrypt(params=params)
        self.assertEqual(
            '52e7fcf91356eca55fc5d52f16f5d777e3521f54e3c570c9bbb7df58fc15add73994e5db42be368de7ebed93c9d4f21f9be7cc453358d734b04a057d0ed3626d',
            result.key)

        with self.assertRaises(TonException):
            params.dk_len = 0
            async_core_client.crypto.scrypt(params=params)

    def test_chacha20(self):
        key = '01' * 32
        nonce = 'ff' * 12
        data = base64.b64encode(b'Message').decode()

        params = ParamsOfChaCha20(data=data, key=key, nonce=nonce)
        encrypted = async_core_client.crypto.chacha20(params=params)
        self.assertEqual('w5QOGsJodQ==', encrypted.data)

        params.data = encrypted.data
        decrypted = async_core_client.crypto.chacha20(params=params)
        self.assertEqual(data, decrypted.data)

    def test_signing_box(self):
        keypair = async_core_client.crypto.generate_random_sign_keys()

        # Create handle
        signing_box = async_core_client.crypto.get_signing_box(params=keypair)
        self.assertIsInstance(signing_box.handle, int)

        # Get public key from box
        result = async_core_client.crypto.signing_box_get_public_key(
            params=signing_box)
        self.assertEqual(keypair.public, result.pubkey)

        # Sign with box
        message = base64.b64encode(b'Sign with box').decode()

        params = ParamsOfSigningBoxSign(
            signing_box=signing_box.handle, unsigned=message)
        box_result = async_core_client.crypto.signing_box_sign(params=params)

        sign_params = ParamsOfSign(unsigned=message, keys=keypair)
        sign_result = async_core_client.crypto.sign(params=sign_params)
        self.assertEqual(sign_result.signature, box_result.signature)

        # Remove signing box
        async_core_client.crypto.remove_signing_box(params=signing_box)

    def test_register_signing_box(self):
        from concurrent.futures import ThreadPoolExecutor
        keys = async_core_client.crypto.generate_random_sign_keys()
        keys_box_handle = async_core_client.crypto.get_signing_box(params=keys)

        def __callback(response_data, *args):
            request = ParamsOfAppRequest(**response_data)
            box_params = ParamsOfAppSigningBox.from_dict(
                data=request.request_data)
            box_result = None

            if isinstance(box_params, ParamsOfAppSigningBox.GetPublicKey):
                # Run method and wait for result
                with ThreadPoolExecutor() as executor:
                    future = executor.submit(
                        async_core_client.crypto.signing_box_get_public_key,
                        params=keys_box_handle)
                    _result = future.result()

                # Resolve params
                box_result = ResultOfAppSigningBox.GetPublicKey(
                    public_key=_result.pubkey)
            if isinstance(box_params, ParamsOfAppSigningBox.Sign):
                # Run method and wait for result
                params = ParamsOfSigningBoxSign(
                    signing_box=keys_box_handle.handle,
                    unsigned=box_params.unsigned)
                with ThreadPoolExecutor() as executor:
                    future = executor.submit(
                        async_core_client.crypto.signing_box_sign,
                        params=params)
                    _result = future.result()

                # Resolve params
                box_result = ResultOfAppSigningBox.Sign(
                    signature=_result.signature)

            # Create resolve app request params
            request_result = AppRequestResult.Ok(
                result=box_result.dict)
            resolve_params = ParamsOfResolveAppRequest(
                app_request_id=request.app_request_id,
                result=request_result)
            with ThreadPoolExecutor() as executor:
                future = executor.submit(
                    async_core_client.resolve_app_request,
                    params=resolve_params)
                future.result()

        # Get external signing box
        external_box = async_core_client.crypto.register_signing_box(
            callback=__callback)

        # Request box public key
        box_pubkey = async_core_client.crypto.signing_box_get_public_key(
            params=external_box)
        self.assertEqual(keys.public, box_pubkey.pubkey)

        # Get signature from signing box
        unsigned = base64.b64encode(b'Test Message').decode()
        sign_params = ParamsOfSigningBoxSign(
            signing_box=external_box.handle, unsigned=unsigned)
        box_sign = async_core_client.crypto.signing_box_sign(
            params=sign_params)

        # Get signature by keys
        sign_params = ParamsOfSign(unsigned=unsigned, keys=keys)
        keys_sign = async_core_client.crypto.sign(params=sign_params)

        self.assertEqual(keys_sign.signature, box_sign.signature)

        async_core_client.crypto.remove_signing_box(params=external_box)


class TestTonCryptoSyncCore(unittest.TestCase):
    """ Sync core is not recommended to use, so make just a couple of tests """
    def test_sha256(self):
        data = base64.b64encode('TON is our future'.encode()).decode()
        params = ParamsOfHash(data=data)
        result = sync_core_client.crypto.sha256(params=params)
        self.assertEqual(
            '1e7fd5ec201652b5375e5edf3e86d0513394d2c2004dd506415abf0578261951',
            result.hash)

        data = base64.b64encode(
            bytes.fromhex('544f4e206973206f757220667574757265')).decode()
        params.data = data
        result = sync_core_client.crypto.sha256(params=params)
        self.assertEqual(
            '1e7fd5ec201652b5375e5edf3e86d0513394d2c2004dd506415abf0578261951',
            result.hash)

        data = 'VE9OIGlzIG91ciBmdXR1cmU='
        params.data = data
        result = sync_core_client.crypto.sha256(params=params)
        self.assertEqual(
            '1e7fd5ec201652b5375e5edf3e86d0513394d2c2004dd506415abf0578261951',
            result.hash)

    def test_sha512(self):
        data = base64.b64encode('TON is our future'.encode()).decode()
        params = ParamsOfHash(data=data)
        result = sync_core_client.crypto.sha512(params=params)
        self.assertEqual(
            '4c52dd4cefc68319bac5e97c1f0d18ae8194fb0dd8d9e090ba8376834a0756175a9a736d1e69cb1a58d25c3d554b02a2b8ed9c3ae5cbeeccc3277746a363a434',
            result.hash)

    def test_scrypt(self):
        password = base64.b64encode('Test Password'.encode()).decode()
        salt = base64.b64encode('Test Salt'.encode()).decode()

        params = ParamsOfScrypt(
            password=password, salt=salt, log_n=10, r=8, p=16, dk_len=64)
        result = sync_core_client.crypto.scrypt(params=params)
        self.assertEqual(
            '52e7fcf91356eca55fc5d52f16f5d777e3521f54e3c570c9bbb7df58fc15add73994e5db42be368de7ebed93c9d4f21f9be7cc453358d734b04a057d0ed3626d',
            result.key)

        with self.assertRaises(TonException):
            params.dk_len = 0
            sync_core_client.crypto.scrypt(params=params)



import base64
import os
import time
from datetime import datetime

import unittest

from tonclient.client import TonClient
from tonclient.errors import TonException
from tonclient.test.helpers import async_core_client, sync_core_client, \
    SAMPLES_DIR, send_grams, async_custom_client, CUSTOM_BASE_URL, \
    GIVER_ADDRESS
from tonclient.types import ParamsOfQueryCollection, OrderBy, SortDirection, \
    ParamsOfWaitForCollection, ParamsOfQuery, ParamsOfSubscribeCollection, \
    SubscriptionResponseType, ResultOfSubscription, ClientError, Abi, \
    ParamsOfEncodeMessage, Signer, DeploySet, CallSet, ParamsOfProcessMessage, \
    ClientConfig, ParamsOfFindLastShardBlock


class TestTonNetAsyncCore(unittest.TestCase):
    def test_query_collection(self):
        q_params = ParamsOfQueryCollection(
            collection='blocks_signatures', result='id', limit=1)
        result = async_core_client.net.query_collection(params=q_params)
        self.assertGreater(len(result.result), 0)

        q_params = ParamsOfQueryCollection(
            collection='accounts', result='id balance', limit=5)
        result = async_core_client.net.query_collection(params=q_params)
        self.assertEqual(5, len(result.result))

        q_params = ParamsOfQueryCollection(
            collection='messages', result='body created_at', limit=10,
            filter={'created_at': {'gt': 1562342740}},
            order=[OrderBy(path='created_at', direction=SortDirection.ASC)])
        result = async_core_client.net.query_collection(params=q_params)
        self.assertGreater(result.result[0]['created_at'], 1562342740)

        with self.assertRaises(TonException):
            q_params = ParamsOfQueryCollection(
                collection='messages', result='id balance')
            async_core_client.net.query_collection(params=q_params)

    def test_wait_for_collection(self):
        now = int(datetime.now().timestamp())
        q_params = ParamsOfWaitForCollection(
            collection='transactions', result='id now',
            filter={'now': {'gt': now}})
        result = async_core_client.net.wait_for_collection(params=q_params)
        self.assertGreater(result.result['now'], now)

        with self.assertRaises(TonException):
            q_params = ParamsOfWaitForCollection(
                collection='transactions', result='', timeout=1)
            async_core_client.net.wait_for_collection(params=q_params)

    def test_subscribe_collection(self):
        results = []

        def __callback(response_data, response_type, *args):
            if response_type == SubscriptionResponseType.OK:
                result = ResultOfSubscription(**response_data)
                results.append(result.result)
            if response_type == SubscriptionResponseType.ERROR:
                raise TonException(error=ClientError(**response_data))

        now = int(datetime.now().timestamp())
        q_params = ParamsOfSubscribeCollection(
            collection='messages', result='created_at',
            filter={'created_at': {'gt': now}})
        subscription = async_core_client.net.subscribe_collection(
            params=q_params, callback=__callback)

        while True:
            if len(results) > 0 or int(datetime.now().timestamp()) > now + 10:
                async_core_client.net.unsubscribe(params=subscription)
                break
            time.sleep(1)

        self.assertGreater(len(results), 0)

    def test_query(self):
        q_params = ParamsOfQuery(
            query='query($time: Float){messages(filter:{created_at:{ge:$time}}limit:5){id}}',
            variables={'time': int(datetime.now().timestamp()) - 60})
        result = async_core_client.net.query(params=q_params)
        self.assertGreater(len(result.result['data']['messages']), 0)

    def test_suspend_resume(self):
        # Data for contract deployment
        keypair = async_custom_client.crypto.generate_random_sign_keys()
        abi = Abi.from_path(path=os.path.join(SAMPLES_DIR, 'Hello.abi.json'))
        with open(os.path.join(SAMPLES_DIR, 'Hello.tvc'), 'rb') as fp:
            tvc = base64.b64encode(fp.read()).decode()
        signer = Signer.Keys(keys=keypair)
        deploy_set = DeploySet(tvc=tvc)
        call_set = CallSet(function_name='constructor')

        # Prepare deployment params
        encode_params = ParamsOfEncodeMessage(
            abi=abi, signer=signer, deploy_set=deploy_set, call_set=call_set)
        encode = async_custom_client.abi.encode_message(params=encode_params)

        # Subscribe for address deploy transaction status
        transactions = []

        def __callback(response_data, response_type, *args):
            if response_type == SubscriptionResponseType.OK:
                result = ResultOfSubscription(**response_data)
                transactions.append(result.result)
                self.assertEqual(encode.address, result.result['account_addr'])
            if response_type == SubscriptionResponseType.ERROR:
                raise TonException(error=ClientError(**response_data))

        subscribe_params = ParamsOfSubscribeCollection(
            collection='transactions', result='id account_addr',
            filter={'account_addr': {'eq': encode.address}, 'status_name': {'eq': 'Finalized'}})
        subscribe = async_custom_client.net.subscribe_collection(
            params=subscribe_params, callback=__callback)

        # Send grams to new account to create first transaction
        send_grams(address=encode.address)
        # Give some time for subscription to receive all data
        time.sleep(2)

        # Suspend subscription
        async_custom_client.net.suspend()
        time.sleep(2)  # Wait a bit for suspend

        # Deploy to create second transaction.
        # Use another client, because of error: Fetch first block failed:
        # Can not use network module since it is suspended
        second_config = ClientConfig()
        second_config.network.server_address = CUSTOM_BASE_URL
        second_client = TonClient(config=second_config)

        process_params = ParamsOfProcessMessage(
            message_encode_params=encode_params, send_events=False)
        second_client.processing.process_message(params=process_params)
        second_client.destroy_context()

        # Check that second transaction is not received when
        # subscription suspended
        self.assertEqual(1, len(transactions))

        # Resume subscription
        async_custom_client.net.resume()
        time.sleep(2)  # Wait a bit for resume

        # Run contract function to create third transaction
        call_set = CallSet(function_name='touch')
        encode_params = ParamsOfEncodeMessage(
            abi=abi, signer=signer, address=encode.address, call_set=call_set)
        process_params = ParamsOfProcessMessage(
            message_encode_params=encode_params, send_events=False)
        async_custom_client.processing.process_message(params=process_params)

        # Give some time for subscription to receive all data
        time.sleep(2)

        # Check that third transaction is now received after resume
        self.assertEqual(2, len(transactions))
        self.assertNotEqual(transactions[0]['id'], transactions[1]['id'])

        # Unsubscribe
        async_custom_client.net.unsubscribe(params=subscribe)

    def test_find_last_shard_block(self):
        find_params = ParamsOfFindLastShardBlock(address=GIVER_ADDRESS)
        result = async_core_client.net.find_last_shard_block(
            params=find_params)
        self.assertIsInstance(result.block_id, str)


class TestTonNetSyncCore(unittest.TestCase):
    """ Sync core is not recommended to use, so make just a couple of tests """
    def test_query_collection(self):
        q_params = ParamsOfQueryCollection(
            collection='blocks_signatures', result='id', limit=1)
        result = sync_core_client.net.query_collection(params=q_params)
        self.assertGreater(len(result.result), 0)

        q_params = ParamsOfQueryCollection(
            collection='accounts', result='id balance', limit=5)
        result = sync_core_client.net.query_collection(params=q_params)
        self.assertEqual(5, len(result.result))

        q_params = ParamsOfQueryCollection(
            collection='messages', filter={'created_at': {'gt': 1562342740}},
            result='body created_at', limit=10,
            order=[OrderBy(path='created_at', direction=SortDirection.ASC)])
        result = sync_core_client.net.query_collection(params=q_params)
        self.assertGreater(result.result[0]['created_at'], 1562342740)

        with self.assertRaises(TonException):
            q_params = ParamsOfQueryCollection(
                collection='messages', result='')
            sync_core_client.net.query_collection(params=q_params)

    def test_wait_for_collection(self):
        now = int(datetime.now().timestamp())
        q_params = ParamsOfWaitForCollection(
            collection='transactions', filter={'now': {'gt': now}},
            result='id now')
        result = sync_core_client.net.wait_for_collection(params=q_params)
        self.assertGreater(result.result['now'], now)

        with self.assertRaises(TonException):
            q_params = ParamsOfWaitForCollection(
                collection='transactions', filter={'now': {'gt': now}},
                result='id now', timeout=1)
            sync_core_client.net.wait_for_collection(params=q_params)



from tonclient.decorators import result_as
from tonclient.module import TonModule
from tonclient.types import ParamsOfRunExecutor, ResultOfRunExecutor, \
    ParamsOfRunTvm, ResultOfRunTvm, ParamsOfRunGet, ResultOfRunGet


class TonTvm(TonModule):
    """ Free TON tvm SDK API implementation """
    @result_as(classname=ResultOfRunGet)
    def run_get(self, params: ParamsOfRunGet) -> ResultOfRunGet:
        """
        Executes a get method of FIFT contract.
        Executes a get method of FIFT contract that fulfills the
        smc-guidelines https://test.ton.org/smc-guidelines.txt and returns
        the result data from TVM's stack
        :param params: See `types.ParamsOfRunGet`
        :return: See `types.ResultOfRunGet`
        """
        return self.request(method='tvm.run_get', **params.dict)

    @result_as(classname=ResultOfRunExecutor)
    def run_executor(
            self, params: ParamsOfRunExecutor) -> ResultOfRunExecutor:
        """
        Emulates all the phases of contract execution locally.

        Performs all the phases of contract execution on Transaction Executor -
        the same component that is used on Validator Nodes.

        Can be used for contract debug, to find out the reason of message
        unsuccessful delivery - as Validators just throw away failed
        transactions, here you can catch it.

        Another use case is to estimate fees for message execution.
        Set `AccountForExecutor::Account.unlimited_balance` to true so that
        emulation will not depend on the actual balance.

        One more use case - you can produce the sequence of operations,
        thus emulating the multiple contract calls locally. And so on.

        To get the account boc (bag of cells) - use `net.query` method to
        download it from graphql api (field `boc` of `account`) or generate
        it with `abi.encode_account` method.
        To get the message boc - use `abi.encode_message` or prepare it any
        other way, for instance, with Fift script.

        If you need this emulation to be as precise as possible then specify
        `ParamsOfRunExecutor` parameter.
        If you need to see the aborted transaction as a result, not as an
        error, set `skip_transaction_check` to `true`
        :param params: See `types.ParamsOfRunExecutor`
        :return: `types.ResultOfRunExecutor`
        """
        return self.request(method='tvm.run_executor', **params.dict)

    @result_as(classname=ResultOfRunTvm)
    def run_tvm(self, params: ParamsOfRunTvm) -> ResultOfRunTvm:
        """
        Executes get methods of ABI-compatible contracts.

        Performs only a part of compute phase of transaction execution that is
        used to run get-methods of ABI-compatible contracts.

        If you try to run get methods with `run_executor` you will get an
        error, because it checks `ACCEPT` and exits if there is none, which
        is actually true for get methods.

        To get the account boc (bag of cells) - use `net.query` method to
        download it from graphql api (field `boc` of `account`) or generate
        it with `abi.encode_account` method.
        To get the message boc - use `abi.encode_message` or prepare it any
        other way, for instance, with Fift script.

        Attention! Updated account state is produces as well, but only
        `account_state.storage.state.data` part of the boc is updated.
        :param params: See `types.ParamsOfRunTvm`
        :return: See `types.ResultOfRunTvm`
        """
        return self.request(method='tvm.run_tvm', **params.dict)



import functools
import inspect


def result_as(classname):
    """ Client response decorator """
    def decorator(function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            def _parse(result):
                if hasattr(classname, 'from_dict'):
                    return classname.from_dict(data=result)
                return classname(**result)

            def _sync_response():
                """ Decorate synchronous request response """
                return _parse(result=fn_result)

            async def _async_response():
                """ Decorate asynchronous request response """
                _fn_result = await fn_result
                return _parse(result=_fn_result)

            # Execute function
            fn_result = function(*args, **kwargs)

            # Return decorated result depending on sync/async request
            if inspect.isawaitable(fn_result):
                return _async_response()
            return _sync_response()
        return wrapper
    return decorator



from typing import Callable

from tonclient.module import TonModule
from tonclient.decorators import result_as
from tonclient.types import KeyPair, ParamsOfFactorize, ResultOfFactorize, \
    ParamsOfModularPower, ResultOfModularPower, ParamsOfTonCrc16, \
    ResultOfTonCrc16, ParamsOfGenerateRandomBytes, \
    ParamsOfConvertPublicKeyToTonSafeFormat, ResultOfGenerateRandomBytes, \
    ResultOfConvertPublicKeyToTonSafeFormat, ParamsOfSign, ResultOfSign, \
    ParamsOfVerifySignature, ResultOfVerifySignature, ParamsOfHash, \
    ResultOfHash, ParamsOfScrypt, ResultOfScrypt, \
    ParamsOfNaclSignKeyPairFromSecret, ParamsOfNaclSign, ResultOfNaclSign, \
    ParamsOfNaclSignOpen, ResultOfNaclSignOpen, ResultOfNaclSignDetached, \
    ParamsOfNaclBoxKeyPairFromSecret, ParamsOfNaclBox, ResultOfNaclBox, \
    ParamsOfNaclBoxOpen, ResultOfNaclBoxOpen, ParamsOfNaclSecretBox, \
    ParamsOfNaclSecretBoxOpen, ParamsOfMnemonicWords, ResultOfMnemonicWords, \
    ParamsOfMnemonicFromRandom, ResultOfMnemonicFromRandom, \
    ParamsOfMnemonicFromEntropy, ResultOfMnemonicFromEntropy, \
    ParamsOfMnemonicVerify, ResultOfMnemonicVerify, \
    ParamsOfMnemonicDeriveSignKeys, ParamsOfHDKeyXPrvFromMnemonic, \
    ResultOfHDKeyXPrvFromMnemonic, ParamsOfHDKeyDeriveFromXPrv, \
    ResultOfHDKeyDeriveFromXPrv, ParamsOfHDKeyDeriveFromXPrvPath, \
    ResultOfHDKeyDeriveFromXPrvPath, ParamsOfHDKeySecretFromXPrv, \
    ResultOfHDKeySecretFromXPrv, ParamsOfHDKeyPublicFromXPrv, \
    ResultOfHDKeyPublicFromXPrv, ParamsOfChaCha20, ResultOfChaCha20, \
    RegisteredSigningBox, ResultOfSigningBoxGetPublicKey, \
    ParamsOfSigningBoxSign, ResultOfSigningBoxSign


class TonCrypto(TonModule):
    """ Free TON crypto SDK API implementation """
    @result_as(classname=ResultOfHash)
    def sha256(self, params: ParamsOfHash) -> ResultOfHash:
        """
        Calculates SHA256 hash of the specified data
        :param params: See `types.ParamsOfHash`
        :return: See `types.ResultOfHash`
        """
        return self.request(method='crypto.sha256', **params.dict)

    @result_as(classname=ResultOfHash)
    def sha512(self, params: ParamsOfHash) -> ResultOfHash:
        """
        Calculates SHA512 hash of the specified data
        :param params: See `types.ParamsOfHash`
        :return: See `types.ResultOfHash`
        """
        return self.request(method='crypto.sha512', **params.dict)

    @result_as(classname=ResultOfHDKeyXPrvFromMnemonic)
    def hdkey_xprv_from_mnemonic(
            self, params: ParamsOfHDKeyXPrvFromMnemonic
    ) -> ResultOfHDKeyXPrvFromMnemonic:
        """
        Generates an extended master private key that will be the root for all
        the derived keys
        :param params: See `types.ParamsOfHDKeyXPrvFromMnemonic`
        :return: See `types.ResultOfHDKeyXPrvFromMnemonic`
        """
        return self.request(
            method='crypto.hdkey_xprv_from_mnemonic', **params.dict)

    @result_as(classname=ResultOfHDKeySecretFromXPrv)
    def hdkey_secret_from_xprv(
            self, params: ParamsOfHDKeySecretFromXPrv
    ) -> ResultOfHDKeySecretFromXPrv:
        """
        Extracts the private key from the serialized extended private key
        :param params: See `types.ParamsOfHDKeySecretFromXPrv`
        :return: See `types.ResultOfHDKeySecretFromXPrv`
        """
        return self.request(
            method='crypto.hdkey_secret_from_xprv', **params.dict)

    @result_as(classname=ResultOfHDKeyPublicFromXPrv)
    def hdkey_public_from_xprv(
            self, params: ParamsOfHDKeyPublicFromXPrv
    ) -> ResultOfHDKeyPublicFromXPrv:
        """
        Extracts the public key from the serialized extended private key
        :param params: See `types.ParamsOfHDKeyPublicFromXPrv`
        :return: See `types.ResultOfHDKeyPublicFromXPrv`
        """
        return self.request(
            method='crypto.hdkey_public_from_xprv', **params.dict)

    @result_as(classname=ResultOfHDKeyDeriveFromXPrv)
    def hdkey_derive_from_xprv(
            self, params: ParamsOfHDKeyDeriveFromXPrv
    ) -> ResultOfHDKeyDeriveFromXPrv:
        """
        Returns extended private key derived from the specified extended
        private key and child index
        :param params: See `types.ParamsOfHDKeyDeriveFromXPrv`
        :return: See `types.ResultOfHDKeyDeriveFromXPrv`
        """
        return self.request(
            method='crypto.hdkey_derive_from_xprv', **params.dict)

    @result_as(classname=ResultOfHDKeyDeriveFromXPrvPath)
    def hdkey_derive_from_xprv_path(
            self, params: ParamsOfHDKeyDeriveFromXPrvPath
    ) -> ResultOfHDKeyDeriveFromXPrvPath:
        """
        Derives the extended private key from the specified key and path
        :param params: See `types.ParamsOfHDKeyDeriveFromXPrvPath`
        :return: See `types.ResultOfHDKeyDeriveFromXPrvPath`
        """
        return self.request(
            method='crypto.hdkey_derive_from_xprv_path', **params.dict)

    @result_as(classname=ResultOfConvertPublicKeyToTonSafeFormat)
    def convert_public_key_to_ton_safe_format(
            self, params: ParamsOfConvertPublicKeyToTonSafeFormat
    ) -> ResultOfConvertPublicKeyToTonSafeFormat:
        """
        Converts public key to ton safe_format
        :param params: See `types.ParamsOfConvertPublicKeyToTonSafeFormat`
        :return: See `types.ResultOfConvertPublicKeyToTonSafeFormat`
        """
        return self.request(
            method='crypto.convert_public_key_to_ton_safe_format',
            **params.dict)

    @result_as(classname=KeyPair)
    def generate_random_sign_keys(self) -> KeyPair:
        """
        Generates random ed25519 key pair
        :return: See `types.KeyPair`
        """
        return self.request(method='crypto.generate_random_sign_keys')

    @result_as(classname=ResultOfSign)
    def sign(self, params: ParamsOfSign) -> ResultOfSign:
        """
        Signs a data using the provided keys
        :param params: See `types.ParamsOfSign`
        :return: See `types.ResultOfSign`
        """
        return self.request(method='crypto.sign', **params.dict)

    @result_as(classname=ResultOfVerifySignature)
    def verify_signature(
            self, params: ParamsOfVerifySignature) -> ResultOfVerifySignature:
        """
        Verifies signed data using the provided public key.
        Raises error if verification is failed
        :param params: See `types.ParamsOfVerifySignature`
        :return: See `types.ResultOfVerifySignature`
        """
        return self.request(method='crypto.verify_signature', **params.dict)

    @result_as(classname=ResultOfModularPower)
    def modular_power(
            self, params: ParamsOfModularPower) -> ResultOfModularPower:
        """
        Performs modular exponentiation for big integers (`base`^`exponent`
        mod `modulus`).
        See [https://en.wikipedia.org/wiki/Modular_exponentiation]
        :param params: See `types.ParamsOfModularPower`
        :return: See `types.ResultOfModularPower`
        """
        return self.request(method='crypto.modular_power', **params.dict)

    @result_as(classname=ResultOfFactorize)
    def factorize(self, params: ParamsOfFactorize) -> ResultOfFactorize:
        """
        Performs prime factorization – decomposition of a composite number
        into a product of smaller prime integers (factors).
        See [https://en.wikipedia.org/wiki/Integer_factorization]
        :param params: See `types.ParamsOfFactorize`
        :return: See `types.ResultOfFactorize`
        """
        return self.request(method='crypto.factorize', **params.dict)

    @result_as(classname=ResultOfTonCrc16)
    def ton_crc16(self, params: ParamsOfTonCrc16) -> ResultOfTonCrc16:
        """
        Calculates CRC16 using TON algorithm
        :param params: See `types.ParamsOfTonCrc16`
        :return: See `types.ResultOfTonCrc16`
        """
        return self.request(method='crypto.ton_crc16', **params.dict)

    @result_as(classname=ResultOfGenerateRandomBytes)
    def generate_random_bytes(
            self, params: ParamsOfGenerateRandomBytes
    ) -> ResultOfGenerateRandomBytes:
        """
        Generates random byte array of the specified length and returns it
        in `base64` format
        :param params: See `types.ParamsOfGenerateRandomBytes`
        :return: See `types.ResultOfGenerateRandomBytes`
        """
        return self.request(
            method='crypto.generate_random_bytes', **params.dict)

    @result_as(classname=ResultOfMnemonicWords)
    def mnemonic_words(
            self, params: ParamsOfMnemonicWords) -> ResultOfMnemonicWords:
        """
        Prints the list of words from the specified dictionary
        :param params: See `types.ParamsOfMnemonicWords`
        :return: See `types.ResultOfMnemonicWords`
        """
        return self.request(method='crypto.mnemonic_words', **params.dict)

    @result_as(classname=ResultOfMnemonicFromRandom)
    def mnemonic_from_random(
            self, params: ParamsOfMnemonicFromRandom
    ) -> ResultOfMnemonicFromRandom:
        """
        Generates a random mnemonic from the specified dictionary
        and word count
        :param params: See `types.ParamsOfMnemonicFromRandom`
        :return: See `types.ResultOfMnemonicFromRandom`
        """
        return self.request(
            method='crypto.mnemonic_from_random', **params.dict)

    @result_as(classname=ResultOfMnemonicFromEntropy)
    def mnemonic_from_entropy(
            self, params: ParamsOfMnemonicFromEntropy
    ) -> ResultOfMnemonicFromEntropy:
        """
        Generates mnemonic from pre-generated entropy
        :param params: See `types.ParamsOfMnemonicFromEntropy`
        :return: See `types.ResultOfMnemonicFromEntropy`
        """
        return self.request(
            method='crypto.mnemonic_from_entropy', **params.dict)

    @result_as(classname=ResultOfMnemonicVerify)
    def mnemonic_verify(
            self, params: ParamsOfMnemonicVerify) -> ResultOfMnemonicVerify:
        """
        The phrase supplied will be checked for word length and validated
        according to the checksum specified in BIP0039
        :param params: See `types.ParamsOfMnemonicVerify`
        :return: See `types.ResultOfMnemonicVerify`
        """
        return self.request(method='crypto.mnemonic_verify', **params.dict)

    @result_as(classname=KeyPair)
    def mnemonic_derive_sign_keys(
            self, params: ParamsOfMnemonicDeriveSignKeys) -> KeyPair:
        """
        Validates the seed phrase, generates master key and then derives
        the key pair from the master key and the specified path
        :param params: See `types.ParamsOfMnemonicDeriveSignKeys`
        :return: See `types.KeyPair`
        """
        return self.request(
            method='crypto.mnemonic_derive_sign_keys', **params.dict)

    @result_as(classname=KeyPair)
    def nacl_sign_keypair_from_secret_key(
            self, params: ParamsOfNaclSignKeyPairFromSecret) -> KeyPair:
        """
        Generates a key pair for signing from the secret key
        :param params: See `types.ParamsOfNaclSignKeyPairFromSecret`
        :return: See `types.KeyPair`
        """
        return self.request(
            method='crypto.nacl_sign_keypair_from_secret_key', **params.dict)

    @result_as(classname=ResultOfNaclSign)
    def nacl_sign(self, params: ParamsOfNaclSign) -> ResultOfNaclSign:
        """
        Signs data using the signer's secret key
        :param params: See `types.ParamsOfNaclSign`
        :return: See `types.ResultOfNaclSign`
        """
        return self.request(method='crypto.nacl_sign', **params.dict)

    @result_as(classname=ResultOfNaclSignDetached)
    def nacl_sign_detached(
            self, params: ParamsOfNaclSign) -> ResultOfNaclSignDetached:
        """
        Signs the message using the secret key and returns a signature.
        Signs the message unsigned using the secret key `secret` and returns a
        signature `signature`
        :param params: See `types.ParamsOfNaclSign`
        :return: See `types.ResultOfNaclSignDetached`
        """
        return self.request(method='crypto.nacl_sign_detached', **params.dict)

    @result_as(classname=ResultOfNaclSignOpen)
    def nacl_sign_open(
            self, params: ParamsOfNaclSignOpen) -> ResultOfNaclSignOpen:
        """
        Verifies the signature and returns the unsigned message.
        Verifies the signature in `signed` using the signer's public key
        `public` and returns the message `unsigned`.
        If the signature fails verification, raises an exception
        :param params: See `types.ParamsOfNaclSignOpen`
        :return: See `types.ResultOfNaclSignOpen`
        """
        return self.request(method='crypto.nacl_sign_open', **params.dict)

    @result_as(classname=KeyPair)
    def nacl_box_keypair(self) -> KeyPair:
        """ Generates a random NaCl key pair """
        return self.request(method='crypto.nacl_box_keypair')

    @result_as(classname=KeyPair)
    def nacl_box_keypair_from_secret_key(
            self, params: ParamsOfNaclBoxKeyPairFromSecret) -> KeyPair:
        """
        Generates key pair from a secret key
        :param params: See `types.ParamsOfNaclBoxKeyPairFromSecret`
        :return: See `types.KeyPair`
        """
        return self.request(
            method='crypto.nacl_box_keypair_from_secret_key', **params.dict)

    @result_as(classname=ResultOfNaclBox)
    def nacl_box(self, params: ParamsOfNaclBox) -> ResultOfNaclBox:
        """
        Public key authenticated encryption.
        Encrypt and authenticate a message using the senders secret key,
        the receivers public key, and a nonce
        :param params: See `types.ParamsOfNaclBox`
        :return: See `types.ResultOfNaclBox`
        """
        return self.request(method='crypto.nacl_box', **params.dict)

    @result_as(classname=ResultOfNaclBoxOpen)
    def nacl_box_open(
            self, params: ParamsOfNaclBoxOpen) -> ResultOfNaclBoxOpen:
        """
        Decrypt and verify the cipher text using the receivers secret key,
        the senders public key, and the nonce
        :param params: See `types.ParamsOfNaclBoxOpen`
        :return: See `types.ResultOfNaclBoxOpen`
        """
        return self.request(method='crypto.nacl_box_open', **params.dict)

    @result_as(classname=ResultOfNaclBox)
    def nacl_secret_box(
            self, params: ParamsOfNaclSecretBox) -> ResultOfNaclBox:
        """
        Encrypt and authenticate message using nonce and secret key
        :param params: See `types.ParamsOfNaclSecretBox`
        :return: See `types.ResultOfNaclBox`
        """
        return self.request(method='crypto.nacl_secret_box', **params.dict)

    @result_as(classname=ResultOfNaclBoxOpen)
    def nacl_secret_box_open(
            self, params: ParamsOfNaclSecretBoxOpen) -> ResultOfNaclBoxOpen:
        """
        Decrypts and verifies cipher text using `nonce` and secret `key`
        :param params: See `types.ParamsOfNaclSecretBoxOpen`
        :return: See `types.ResultOfNaclBoxOpen`
        """
        return self.request(
            method='crypto.nacl_secret_box_open', **params.dict)

    @result_as(classname=ResultOfScrypt)
    def scrypt(self, params: ParamsOfScrypt) -> ResultOfScrypt:
        """
        Derives key from `password` and `key` using `scrypt` algorithm.
        See [https://en.wikipedia.org/wiki/Scrypt].
        :param params: See `types.ParamsOfScrypt`
        :return: See `types.ResultOfScrypt`
        """
        return self.request(method='crypto.scrypt', **params.dict)

    @result_as(classname=ResultOfChaCha20)
    def chacha20(self, params: ParamsOfChaCha20) -> ResultOfChaCha20:
        """
        Performs symmetric `chacha20` encryption
        :param params: See `types.ParamsOfChaCha20`
        :return: See `types.ResultOfChaCha20`
        """
        return self.request(method='crypto.chacha20', **params.dict)

    @result_as(classname=RegisteredSigningBox)
    def register_signing_box(self, callback: Callable) -> RegisteredSigningBox:
        """
        Register an application implemented signing box
        :param callback: Callback to send events to
        :return: See `types.RegisteredSigningBox`
        """
        return self.request(
            method='crypto.register_signing_box', callback=callback)

    @result_as(classname=RegisteredSigningBox)
    def get_signing_box(self, params: KeyPair) -> RegisteredSigningBox:
        """
        Creates a default signing box implementation
        :param params: See `types.KeyPair`
        :return: See `types.RegisteredSigningBox`
        """
        return self.request(method='crypto.get_signing_box', **params.dict)

    @result_as(classname=ResultOfSigningBoxGetPublicKey)
    def signing_box_get_public_key(
            self, params: RegisteredSigningBox
    ) -> ResultOfSigningBoxGetPublicKey:
        """
        Returns public key of signing key pair
        :param params: See `types.RegisteredSigningBox`
        :return: See `types.ResultOfSigningBoxGetPublicKey`
        """
        return self.request(
            method='crypto.signing_box_get_public_key', **params.dict)

    @result_as(classname=ResultOfSigningBoxSign)
    def signing_box_sign(
            self, params: ParamsOfSigningBoxSign) -> ResultOfSigningBoxSign:
        """
        Returns signed user data
        :param params: See `types.ParamsOfSigningBoxSign`
        :return: See `types.ResultOfSigningBoxSign`
        """
        return self.request(method='crypto.signing_box_sign', **params.dict)

    def remove_signing_box(self, params: RegisteredSigningBox):
        """
        Removes signing box from SDK
        :param params: See `types.RegisteredSigningBox`
        :return:
        """
        return self.request(method='crypto.remove_signing_box', **params.dict)



from tonclient.decorators import result_as
from tonclient.module import TonModule
from tonclient.types import ParamsOfConvertAddress, ResultOfConvertAddress


class TonUtils(TonModule):
    """ Free TON utils SDK API implementation """
    @result_as(classname=ResultOfConvertAddress)
    def convert_address(
            self, params: ParamsOfConvertAddress) -> ResultOfConvertAddress:
        """
        Converts address from any TON format to any TON format
        :param params: See `types.ParamsOfConvertAddress`
        :return: See `types.ResultOfConvertAddress`
        """
        return self.request(method='utils.convert_address', **params.dict)



import asyncio
import logging
import os

import json
import threading
import time
from typing import Any, Dict, Union, Callable

from tonclient.bindings.lib import tc_request, tc_request_sync, \
    tc_read_string, tc_destroy_string
from tonclient.bindings.types import TCStringData, TCResponseHandler, \
    TCResponseType, TCSyncResponseData
from tonclient.errors import TonException
from tonclient.types import ClientError


class ResolverThread(threading.Thread):
    """ Thread to resolve async core response """
    def __init__(
            self, resolved: bool = False, result: Any = None, *args, **kwargs):
        super(ResolverThread, self).__init__(*args, **kwargs)
        self.resolved = resolved
        self.result = result
        self.exception = None

    def run(self) -> None:
        while not self.resolved and not self.exception:
            time.sleep(0.001)
        return self.result


class TonModule(object):
    """
    Base TON Module class.
    All modules, such as 'crypto', 'contracts', etc. should be inherited
    from this class.
    """
    _async_response_map = {}

    def __init__(self, client):
        self._client = client

    def request(
            self, method: str, callback: Callable = None,
            params_or_str: Union[str, Dict[str, Any]] = None, **kwargs) -> Any:
        """ Perform core request """
        # Prepare request params
        request_params = self._prepare_params(params_or_str, **kwargs)
        kwargs = {
            'method': method,
            'request_params': request_params
        }

        # Make sync or async core/client request
        if self._client.is_core_async:
            kwargs.update({'callback': callback})
            if self._client.is_async:
                return self._async_core_request_future(**kwargs)
            return self._async_core_request(**kwargs)
        return self._sync_core_request(**kwargs)

    def _sync_core_request(self, method: str, request_params: str) -> Any:
        """ Perform core synchronous request """
        # Make sync request, get response pointer and parse it
        response_ptr = tc_request_sync(
            ctx=self._client.ctx, method=method, params_json=request_params)
        response = TCSyncResponseData(tc_read_string(string=response_ptr))

        # Copy response data and destroy response pointer
        is_success, result, error = (
            response.is_success, response.result, response.error)
        tc_destroy_string(string=response_ptr)

        if not is_success:
            raise TonException(error=ClientError(**error))

        return result

    def _async_core_request(
            self, method: str, request_params: str, callback: Callable) -> Any:
        """ Perform core asynchronous request """
        # Generate request id
        request_id = self._generate_request_id()

        # Create resolving thread and start it
        thread = ResolverThread(daemon=True)
        thread.start()

        # Set response map data
        self._async_response_map[request_id] = {
            'is_async': False,
            'callback': callback,
            'thread': thread
        }

        # Execute core request
        tc_request(
            ctx=self._client.ctx, method=method, request_id=request_id,
            params_json=request_params,
            response_handler=self._async_response_handler)

        # Wait for resolver thread
        thread.join()
        if thread.exception:
            raise thread.exception

        return thread.result

    async def _async_core_request_future(
            self, method: str, request_params: str, callback: Callable
    ) -> Any:
        """ Perform core asynchronous request """
        # Generate request id
        request_id = self._generate_request_id()

        # Get event loop and create future
        loop = asyncio.get_event_loop()
        future = loop.create_future()

        # Set response map data
        self._async_response_map[request_id] = {
            'is_async': True,
            'callback': callback,
            'loop': loop,
            'future': future
        }

        # Execute core request
        tc_request(
            ctx=self._client.ctx, method=method, request_id=request_id,
            params_json=request_params,
            response_handler=self._async_response_handler)

        return await future

    def _generate_request_id(self, size: int = 3) -> int:
        """
        Generate random int of size in bytes
        :param size: Length in bytes
        :return:
        """
        while True:
            request_id = int.from_bytes(os.urandom(size), 'big')
            if request_id in list(self._async_response_map.keys()):
                continue
            return request_id

    @staticmethod
    def _prepare_params(params_or_str, **kwargs) -> str:
        """ Prepare params to pass to request """
        if isinstance(params_or_str, dict):
            params_or_str = {**params_or_str, **kwargs}
        elif params_or_str is None:
            params_or_str = kwargs or {}

        return json.dumps(params_or_str)

    @staticmethod
    @TCResponseHandler
    def _async_response_handler(
            request_id: int, response_data: TCStringData,
            response_type: int, finished: bool):
        """ Core response handler """
        logging.debug(
            f'Request: {request_id}; Response: {response_data.json}; '
            f'Response type: {response_type}; Finished: {finished}')

        request = TonModule._async_response_map.get(request_id)
        if not request:
            return
        if finished:
            del TonModule._async_response_map[request_id]

        if response_type == TCResponseType.Success:
            # Check if client is asyncio or common
            if request['is_async']:
                request['loop'].call_soon_threadsafe(
                    request['future'].set_result, response_data.json)
            else:
                request['thread'].result = response_data.json
                request['thread'].resolved = True
            return

        if response_type == TCResponseType.Error:
            exception = TonException(error=ClientError(**response_data.json))
            # Check if client is asyncio or common
            if request['is_async']:
                request['loop'].call_soon_threadsafe(
                    request['future'].set_exception, exception)
            else:
                request['thread'].exception = exception
            return

        if request['callback'] and response_data.json:
            args = [response_data.json, response_type, request.get('loop')]
            request['callback'](*args)



import unittest

from tonclient.errors import TonException
from tonclient.test.helpers import async_core_client, sync_core_client
from tonclient.types import AddressStringFormat, ParamsOfConvertAddress


class TestTonUtilsAsyncCore(unittest.TestCase):
    def test_convert_address(self):
        account_id = 'fcb91a3a3816d0f7b8c2c76108b8a9bc5a6b7a55bd79f8ab101c52db29232260'
        hex_ = '-1:fcb91a3a3816d0f7b8c2c76108b8a9bc5a6b7a55bd79f8ab101c52db29232260'
        hex_workchain0 = '0:fcb91a3a3816d0f7b8c2c76108b8a9bc5a6b7a55bd79f8ab101c52db29232260'
        base64 = 'Uf/8uRo6OBbQ97jCx2EIuKm8Wmt6Vb15+KsQHFLbKSMiYG+9'
        base64url = 'kf_8uRo6OBbQ97jCx2EIuKm8Wmt6Vb15-KsQHFLbKSMiYIny'

        convert_params = ParamsOfConvertAddress(
            address=account_id, output_format=AddressStringFormat.Hex())
        converted = async_core_client.utils.convert_address(
            params=convert_params)
        self.assertEqual(hex_workchain0, converted.address)

        convert_params = ParamsOfConvertAddress(
            address=converted.address,
            output_format=AddressStringFormat.AccountId())
        converted = async_core_client.utils.convert_address(
            params=convert_params)
        self.assertEqual(account_id, converted.address)

        convert_params = ParamsOfConvertAddress(
            address=hex_,
            output_format=AddressStringFormat.Base64(
                url=False, test=False, bounce=False))
        converted = async_core_client.utils.convert_address(
            params=convert_params)
        self.assertEqual(base64, converted.address)

        convert_params = ParamsOfConvertAddress(
            address=base64,
            output_format=AddressStringFormat.Base64(
                url=True, test=True, bounce=True))
        converted = async_core_client.utils.convert_address(
            params=convert_params)
        self.assertEqual(base64url, converted.address)

        convert_params = ParamsOfConvertAddress(
            address=base64url,
            output_format=AddressStringFormat.Hex())
        converted = async_core_client.utils.convert_address(
            params=convert_params)
        self.assertEqual(hex_, converted.address)

        with self.assertRaises(TonException):
            convert_params = ParamsOfConvertAddress(
                address='-1:00',
                output_format=AddressStringFormat.Hex())
            async_core_client.utils.convert_address(params=convert_params)


class TestTonUtilsSyncCore(unittest.TestCase):
    """ Sync core is not recommended to use """
    def test_convert_address(self):
        account_id = 'fcb91a3a3816d0f7b8c2c76108b8a9bc5a6b7a55bd79f8ab101c52db29232260'
        hex_ = '-1:fcb91a3a3816d0f7b8c2c76108b8a9bc5a6b7a55bd79f8ab101c52db29232260'
        hex_workchain0 = '0:fcb91a3a3816d0f7b8c2c76108b8a9bc5a6b7a55bd79f8ab101c52db29232260'
        base64 = 'Uf/8uRo6OBbQ97jCx2EIuKm8Wmt6Vb15+KsQHFLbKSMiYG+9'
        base64url = 'kf_8uRo6OBbQ97jCx2EIuKm8Wmt6Vb15-KsQHFLbKSMiYIny'

        convert_params = ParamsOfConvertAddress(
            address=account_id, output_format=AddressStringFormat.Hex())
        converted = sync_core_client.utils.convert_address(
            params=convert_params)
        self.assertEqual(hex_workchain0, converted.address)

        convert_params = ParamsOfConvertAddress(
            address=converted.address,
            output_format=AddressStringFormat.AccountId())
        converted = sync_core_client.utils.convert_address(
            params=convert_params)
        self.assertEqual(account_id, converted.address)

        convert_params = ParamsOfConvertAddress(
            address=hex_,
            output_format=AddressStringFormat.Base64(
                url=False, test=False, bounce=False))
        converted = sync_core_client.utils.convert_address(
            params=convert_params)
        self.assertEqual(base64, converted.address)

        convert_params = ParamsOfConvertAddress(
            address=base64,
            output_format=AddressStringFormat.Base64(
                url=True, test=True, bounce=True))
        converted = sync_core_client.utils.convert_address(
            params=convert_params)
        self.assertEqual(base64url, converted.address)

        convert_params = ParamsOfConvertAddress(
            address=base64url,
            output_format=AddressStringFormat.Hex())
        converted = sync_core_client.utils.convert_address(
            params=convert_params)
        self.assertEqual(hex_, converted.address)

        with self.assertRaises(TonException):
            convert_params = ParamsOfConvertAddress(
                address='-1:00',
                output_format=AddressStringFormat.Hex())
            sync_core_client.utils.convert_address(params=convert_params)



from typing import Any

from tonclient.types import ClientError


class TonException(Exception):
    def __init__(self, error: Any):
        if isinstance(error, ClientError):
            error = error.__str__()
        super(TonException, self).__init__(error)



from typing import Callable

from tonclient.decorators import result_as
from tonclient.module import TonModule
from tonclient.types import ParamsOfQuery, ResultOfQuery, \
    ParamsOfQueryCollection, ResultOfQueryCollection, \
    ParamsOfWaitForCollection, ResultOfWaitForCollection, \
    ResultOfSubscribeCollection, ParamsOfSubscribeCollection, \
    ParamsOfFindLastShardBlock, ResultOfFindLastShardBlock


class TonNet(TonModule):
    """ Free TON net SDK API implementation """
    @result_as(classname=ResultOfQueryCollection)
    def query_collection(
            self, params: ParamsOfQueryCollection) -> ResultOfQueryCollection:
        """
        Queries collection data.
        Queries data that satisfies the `filter` conditions, limits the number
        of returned records and orders them. The projection fields are limited
        to result fields
        :param params: See `types.ParamsOfQueryCollection`
        :return: See `types.ResultOfQueryCollection`
        """
        return self.request(method='net.query_collection', **params.dict)

    @result_as(classname=ResultOfWaitForCollection)
    def wait_for_collection(
            self, params: ParamsOfWaitForCollection
    ) -> ResultOfWaitForCollection:
        """
        Returns an object that fulfills the conditions or waits for its
        appearance. Triggers only once.
        If object that satisfies the `filter` conditions already exists -
        returns it immediately. If not - waits for insert/update of data
        within the specified `timeout`, and returns it. The projection fields
        are limited to `result` fields
        :param params: See `types.ParamsOfWaitForCollection`
        :return: See `types.ResultOfWaitForCollection`
        """
        return self.request(method='net.wait_for_collection', **params.dict)

    @result_as(classname=ResultOfSubscribeCollection)
    def subscribe_collection(
            self, params: ParamsOfSubscribeCollection,
            callback: Callable = None
    ) -> ResultOfSubscribeCollection:
        """
        Creates a subscription.
        Triggers for each insert/update of data that satisfies the `filter`
        conditions. The projection fields are limited to `result` fields
        :param params: See `types.ParamsOfSubscribeCollection`
        :param callback: Additional responses handler
        :return:
        """
        return self.request(
            method='net.subscribe_collection', callback=callback,
            **params.dict)

    def unsubscribe(self, params: ResultOfSubscribeCollection):
        """
        Cancels a subscription.
        Cancels a subscription specified by its handle
        :param params: See `types.ResultOfSubscribeCollection`
        """
        return self.request(method='net.unsubscribe', **params.dict)

    @result_as(classname=ResultOfQuery)
    def query(self, params: ParamsOfQuery) -> ResultOfQuery:
        """
        Performs DAppServer GraphQL query
        :param params: See `types.ResultOfQuery`
        :return: See `types.ResultOfQuery`
        """
        return self.request(method='net.query', **params.dict)

    def suspend(self):
        """ Suspends network module to stop any network activity """
        return self.request(method='net.suspend')

    def resume(self):
        """ Resumes network module to enable network activity """
        return self.request(method='net.resume')

    @result_as(classname=ResultOfFindLastShardBlock)
    def find_last_shard_block(
            self, params: ParamsOfFindLastShardBlock
    ) -> ResultOfFindLastShardBlock:
        """
        :param params: See `types.ParamsOfFindLastShardBlock`
        :return: See `types.ResultOfFindLastShardBlock`
        """
        return self.request(method='net.find_last_shard_block', **params.dict)



from tonclient.decorators import result_as
from tonclient.module import TonModule
from tonclient.types import ParamsOfParse, ResultOfParse, \
    ParamsOfParseShardstate, ParamsOfGetBocHash, ResultOfGetBocHash, \
    ParamsOfGetBlockchainConfig, ResultOfGetBlockchainConfig, \
    ParamsOfGetCodeFromTvc, ResultOfGetCodeFromTvc


class TonBoc(TonModule):
    """ Free TON boc SDK API implementation """
    @result_as(classname=ResultOfParse)
    def parse_message(self, params: ParamsOfParse) -> ResultOfParse:
        """
        Parses message boc into a JSON.
        JSON structure is compatible with GraphQL API message object
        :param params: See `types.ParamsOfParse`
        :return: See `types.ResultOfParse`
        """
        return self.request(method='boc.parse_message', **params.dict)

    @result_as(classname=ResultOfParse)
    def parse_transaction(self, params: ParamsOfParse) -> ResultOfParse:
        """
        Parses transaction boc into a JSON.
        JSON structure is compatible with GraphQL API transaction object
        :param params: See `types.ParamsOfParse`
        :return: See `types.ResultOfParse`
        """
        return self.request(method='boc.parse_transaction', **params.dict)

    @result_as(classname=ResultOfParse)
    def parse_account(self, params: ParamsOfParse) -> ResultOfParse:
        """
        Parses account boc into a JSON.
        JSON structure is compatible with GraphQL API account object
        :param params: See `types.ParamsOfParse`
        :return: See `types.ResultOfParse`
        """
        return self.request(method='boc.parse_account', **params.dict)

    @result_as(classname=ResultOfParse)
    def parse_block(self, params: ParamsOfParse) -> ResultOfParse:
        """
        Parses block boc into a JSON.
        JSON structure is compatible with GraphQL API block object
        :param params: See `types.ParamsOfParse`
        :return: See `types.ResultOfParse`
        """
        return self.request(method='boc.parse_block', **params.dict)

    @result_as(classname=ResultOfParse)
    def parse_shardstate(
            self, params: ParamsOfParseShardstate) -> ResultOfParse:
        """
        Parses shardstate boc into a JSON.
        JSON structure is compatible with GraphQL API shardstate object
        :param params: See `ParamsOfParseShardstate`
        :return: See `ResultOfParse`
        """
        return self.request(method='boc.parse_shardstate', **params.dict)

    @result_as(classname=ResultOfGetBocHash)
    def get_boc_hash(self, params: ParamsOfGetBocHash) -> ResultOfGetBocHash:
        """
        Calculates BOC root hash
        :param params: See `ParamsOfGetBocHash`
        :return: See `ResultOfGetBocHash`
        """
        return self.request(method='boc.get_boc_hash', **params.dict)

    @result_as(classname=ResultOfGetBlockchainConfig)
    def get_blockchain_config(
            self, params: ParamsOfGetBlockchainConfig
    ) -> ResultOfGetBlockchainConfig:
        """
        :param params: See `ParamsOfGetBlockchainConfig`
        :return: See `ResultOfGetBlockchainConfig`
        """
        return self.request(
            method='boc.get_blockchain_config', **params.dict)

    @result_as(classname=ResultOfGetCodeFromTvc)
    def get_code_from_tvc(
            self, params: ParamsOfGetCodeFromTvc) -> ResultOfGetCodeFromTvc:
        """
        Extracts code from TVC contract image
        :param params: See `types.ParamsOfGetCodeFromTvc`
        :return: See `types.ResultOfGetCodeFromTvc`
        """
        return self.request(method='boc.get_code_from_tvc', **params.dict)



import base64
import os
import unittest

from tonclient.errors import TonException
from tonclient.test.helpers import SAMPLES_DIR, async_core_client, \
    sync_core_client
from tonclient.types import Abi, KeyPair, DeploySet, CallSet, Signer, \
    MessageSource, StateInitSource, FunctionHeader, ParamsOfEncodeMessageBody, \
    ParamsOfDecodeMessage, MessageBodyType, ParamsOfParse, \
    ParamsOfDecodeMessageBody, ParamsOfEncodeMessage, ParamsOfSign, \
    ParamsOfAttachSignature, ParamsOfEncodeAccount


class TestTonAbiAsyncCore(unittest.TestCase):
    def setUp(self) -> None:
        # Events contract params
        self.events_abi = Abi.from_path(
            path=os.path.join(SAMPLES_DIR, 'Events.abi.json'))
        self.keypair = KeyPair(
            public='4c7c408ff1ddebb8d6405ee979c716a14fdd6cc08124107a61d3c25597099499',
            secret='cc8929d635719612a9478b9cd17675a39cfad52d8959e8a177389b8c0b9122a7')
        with open(os.path.join(SAMPLES_DIR, 'Events.tvc'), 'rb') as fp:
            self.events_tvc = base64.b64encode(fp.read()).decode()
        self.events_time = 1599458364291
        self.events_expire = 1599458404

    def test_decode_message(self):
        message = 'te6ccgEBAwEAvAABRYgAC31qq9KF9Oifst6LU9U6FQSQQRlCSEMo+A3LN5MvphIMAQHhrd/b+MJ5Za+AygBc5qS/dVIPnqxCsM9PvqfVxutK+lnQEKzQoRTLYO6+jfM8TF4841bdNjLQwIDWL4UVFdxIhdMfECP8d3ruNZAXul5xxahT91swIEkEHph08JVlwmUmQAAAXRnJcuDX1XMZBW+LBKACAEAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=='
        params = ParamsOfDecodeMessage(abi=self.events_abi, message=message)
        decoded = async_core_client.abi.decode_message(params=params)
        self.assertEqual(MessageBodyType.INPUT, decoded.body_type)
        self.assertEqual(0, int(decoded.value['id'], 16))
        self.assertEqual(self.events_expire, decoded.header.expire)
        self.assertEqual(self.events_time, decoded.header.time)
        self.assertEqual(self.keypair.public, decoded.header.pubkey)

        message = 'te6ccgEBAQEAVQAApeACvg5/pmQpY4m61HmJ0ne+zjHJu3MNG8rJxUDLbHKBu/AAAAAAAAAMJL6z6ro48sYvAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAABA'
        params = ParamsOfDecodeMessage(abi=self.events_abi, message=message)
        decoded = async_core_client.abi.decode_message(params=params)
        self.assertEqual(MessageBodyType.EVENT, decoded.body_type)
        self.assertEqual(0, int(decoded.value['id'], 16))
        self.assertIsNone(decoded.header)

        message = 'te6ccgEBAQEAVQAApeACvg5/pmQpY4m61HmJ0ne+zjHJu3MNG8rJxUDLbHKBu/AAAAAAAAAMKr6z6rxK3xYJAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAABA'
        params = ParamsOfDecodeMessage(abi=self.events_abi, message=message)
        decoded = async_core_client.abi.decode_message(params=params)
        self.assertEqual(MessageBodyType.OUTPUT, decoded.body_type)
        self.assertEqual(0, int(decoded.value['value0'], 16))
        self.assertIsNone(decoded.header)

        with self.assertRaises(TonException):
            params = ParamsOfDecodeMessage(abi=self.events_abi, message='0x0')
            async_core_client.abi.decode_message(params=params)

    def test_decode_message_body(self):
        message = 'te6ccgEBAwEAvAABRYgAC31qq9KF9Oifst6LU9U6FQSQQRlCSEMo+A3LN5MvphIMAQHhrd/b+MJ5Za+AygBc5qS/dVIPnqxCsM9PvqfVxutK+lnQEKzQoRTLYO6+jfM8TF4841bdNjLQwIDWL4UVFdxIhdMfECP8d3ruNZAXul5xxahT91swIEkEHph08JVlwmUmQAAAXRnJcuDX1XMZBW+LBKACAEAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=='
        boc = async_core_client.boc.parse_message(
            params=ParamsOfParse(boc=message))

        decode_params = ParamsOfDecodeMessageBody(
            abi=self.events_abi, body=boc.parsed['body'], is_internal=False)
        decoded = async_core_client.abi.decode_message_body(
            params=decode_params)
        self.assertEqual(MessageBodyType.INPUT, decoded.body_type)
        self.assertEqual(0, int(decoded.value['id'], 16))
        self.assertEqual(self.events_expire, decoded.header.expire)
        self.assertEqual(self.events_time, decoded.header.time)
        self.assertEqual(self.keypair.public, decoded.header.pubkey)

    def test_encode_message(self):
        deploy_set = DeploySet(tvc=self.events_tvc)
        call_set = CallSet(
            function_name='constructor',
            header=FunctionHeader(
                pubkey=self.keypair.public, time=self.events_time,
                expire=self.events_expire))
        signer = Signer.External(public_key=self.keypair.public)

        # Create unsigned deployment message
        params = ParamsOfEncodeMessage(
            abi=self.events_abi, signer=signer, deploy_set=deploy_set,
            call_set=call_set)
        unsigned = async_core_client.abi.encode_message(params=params)
        self.assertEqual(
            'te6ccgECFwEAA2gAAqeIAAt9aqvShfTon7Lei1PVOhUEkEEZQkhDKPgNyzeTL6YSEZTHxAj/Hd67jWQF7peccWoU/dbMCBJBB6YdPCVZcJlJkAAAF0ZyXLg19VzGRotV8/gGAQEBwAICA88gBQMBAd4EAAPQIABB2mPiBH+O713GsgL3S844tQp+62YECSCD0w6eEqy4TKTMAib/APSkICLAAZL0oOGK7VNYMPShCQcBCvSkIPShCAAAAgEgDAoByP9/Ie1E0CDXScIBjhDT/9M/0wDRf/hh+Gb4Y/hijhj0BXABgED0DvK91wv/+GJw+GNw+GZ/+GHi0wABjh2BAgDXGCD5AQHTAAGU0/8DAZMC+ELiIPhl+RDyqJXTAAHyeuLTPwELAGqOHvhDIbkgnzAg+COBA+iogggbd0Cgud6S+GPggDTyNNjTHwH4I7zyudMfAfAB+EdukvI83gIBIBINAgEgDw4AvbqLVfP/hBbo417UTQINdJwgGOENP/0z/TANF/+GH4Zvhj+GKOGPQFcAGAQPQO8r3XC//4YnD4Y3D4Zn/4YeLe+Ebyc3H4ZtH4APhCyMv/+EPPCz/4Rs8LAMntVH/4Z4AgEgERAA5biABrW/CC3Rwn2omhp/+mf6YBov/ww/DN8Mfwxb30gyupo6H0gb+j8IpA3SRg4b3whXXlwMnwAZGT9ghBkZ8KEZ0aCBAfQAAAAAAAAAAAAAAAAACBni2TAgEB9gBh8IWRl//wh54Wf/CNnhYBk9qo//DPAAxbmTwqLfCC3Rwn2omhp/+mf6YBov/ww/DN8Mfwxb2uG/8rqaOhp/+/o/ABkRe4AAAAAAAAAAAAAAAAIZ4tnwOfI48sYvRDnhf/kuP2AGHwhZGX//CHnhZ/8I2eFgGT2qj/8M8AIBSBYTAQm4t8WCUBQB/PhBbo4T7UTQ0//TP9MA0X/4Yfhm+GP4Yt7XDf+V1NHQ0//f0fgAyIvcAAAAAAAAAAAAAAAAEM8Wz4HPkceWMXohzwv/yXH7AMiL3AAAAAAAAAAAAAAAABDPFs+Bz5JW+LBKIc8L/8lx+wAw+ELIy//4Q88LP/hGzwsAye1UfxUABPhnAHLccCLQ1gIx0gAw3CHHAJLyO+Ah1w0fkvI84VMRkvI74cEEIoIQ/////byxkvI84AHwAfhHbpLyPN4=',
            unsigned.message)
        self.assertEqual(
            'KCGM36iTYuCYynk+Jnemis+mcwi3RFCke95i7l96s4Q=',
            unsigned.data_to_sign)

        # Create detached signature
        sign_params = ParamsOfSign(
            unsigned=unsigned.data_to_sign, keys=self.keypair)
        signature = async_core_client.crypto.sign(params=sign_params)
        self.assertEqual(
            '6272357bccb601db2b821cb0f5f564ab519212d242cf31961fe9a3c50a30b236012618296b4f769355c0e9567cd25b366f3c037435c498c82e5305622adbc70e',
            signature.signature)

        # Attach signature to unsigned message
        attach_params = ParamsOfAttachSignature(
            abi=self.events_abi, public_key=self.keypair.public,
            message=unsigned.message, signature=signature.signature)
        signed = async_core_client.abi.attach_signature(params=attach_params)
        self.assertEqual(
            'te6ccgECGAEAA6wAA0eIAAt9aqvShfTon7Lei1PVOhUEkEEZQkhDKPgNyzeTL6YSEbAHAgEA4bE5Gr3mWwDtlcEOWHr6slWoyQlpIWeYyw/00eKFGFkbAJMMFLWnu0mq4HSrPmktmzeeAboa4kxkFymCsRVt44dTHxAj/Hd67jWQF7peccWoU/dbMCBJBB6YdPCVZcJlJkAAAF0ZyXLg19VzGRotV8/gAQHAAwIDzyAGBAEB3gUAA9AgAEHaY+IEf47vXcayAvdLzji1Cn7rZgQJIIPTDp4SrLhMpMwCJv8A9KQgIsABkvSg4YrtU1gw9KEKCAEK9KQg9KEJAAACASANCwHI/38h7UTQINdJwgGOENP/0z/TANF/+GH4Zvhj+GKOGPQFcAGAQPQO8r3XC//4YnD4Y3D4Zn/4YeLTAAGOHYECANcYIPkBAdMAAZTT/wMBkwL4QuIg+GX5EPKoldMAAfJ64tM/AQwAao4e+EMhuSCfMCD4I4ED6KiCCBt3QKC53pL4Y+CANPI02NMfAfgjvPK50x8B8AH4R26S8jzeAgEgEw4CASAQDwC9uotV8/+EFujjXtRNAg10nCAY4Q0//TP9MA0X/4Yfhm+GP4Yo4Y9AVwAYBA9A7yvdcL//hicPhjcPhmf/hh4t74RvJzcfhm0fgA+ELIy//4Q88LP/hGzwsAye1Uf/hngCASASEQDluIAGtb8ILdHCfaiaGn/6Z/pgGi//DD8M3wx/DFvfSDK6mjofSBv6PwikDdJGDhvfCFdeXAyfABkZP2CEGRnwoRnRoIEB9AAAAAAAAAAAAAAAAAAIGeLZMCAQH2AGHwhZGX//CHnhZ/8I2eFgGT2qj/8M8ADFuZPCot8ILdHCfaiaGn/6Z/pgGi//DD8M3wx/DFva4b/yupo6Gn/7+j8AGRF7gAAAAAAAAAAAAAAAAhni2fA58jjyxi9EOeF/+S4/YAYfCFkZf/8IeeFn/wjZ4WAZPaqP/wzwAgFIFxQBCbi3xYJQFQH8+EFujhPtRNDT/9M/0wDRf/hh+Gb4Y/hi3tcN/5XU0dDT/9/R+ADIi9wAAAAAAAAAAAAAAAAQzxbPgc+Rx5YxeiHPC//JcfsAyIvcAAAAAAAAAAAAAAAAEM8Wz4HPklb4sEohzwv/yXH7ADD4QsjL//hDzws/+EbPCwDJ7VR/FgAE+GcActxwItDWAjHSADDcIccAkvI74CHXDR+S8jzhUxGS8jvhwQQighD////9vLGS8jzgAfAB+EdukvI83g==',
            signed.message)

        # Create initially signed message
        signer = Signer.Keys(keys=self.keypair)
        encode_params = ParamsOfEncodeMessage(
            abi=self.events_abi, signer=signer, deploy_set=deploy_set,
            call_set=call_set)
        signed = async_core_client.abi.encode_message(params=encode_params)
        self.assertEqual(
            'te6ccgECGAEAA6wAA0eIAAt9aqvShfTon7Lei1PVOhUEkEEZQkhDKPgNyzeTL6YSEbAHAgEA4bE5Gr3mWwDtlcEOWHr6slWoyQlpIWeYyw/00eKFGFkbAJMMFLWnu0mq4HSrPmktmzeeAboa4kxkFymCsRVt44dTHxAj/Hd67jWQF7peccWoU/dbMCBJBB6YdPCVZcJlJkAAAF0ZyXLg19VzGRotV8/gAQHAAwIDzyAGBAEB3gUAA9AgAEHaY+IEf47vXcayAvdLzji1Cn7rZgQJIIPTDp4SrLhMpMwCJv8A9KQgIsABkvSg4YrtU1gw9KEKCAEK9KQg9KEJAAACASANCwHI/38h7UTQINdJwgGOENP/0z/TANF/+GH4Zvhj+GKOGPQFcAGAQPQO8r3XC//4YnD4Y3D4Zn/4YeLTAAGOHYECANcYIPkBAdMAAZTT/wMBkwL4QuIg+GX5EPKoldMAAfJ64tM/AQwAao4e+EMhuSCfMCD4I4ED6KiCCBt3QKC53pL4Y+CANPI02NMfAfgjvPK50x8B8AH4R26S8jzeAgEgEw4CASAQDwC9uotV8/+EFujjXtRNAg10nCAY4Q0//TP9MA0X/4Yfhm+GP4Yo4Y9AVwAYBA9A7yvdcL//hicPhjcPhmf/hh4t74RvJzcfhm0fgA+ELIy//4Q88LP/hGzwsAye1Uf/hngCASASEQDluIAGtb8ILdHCfaiaGn/6Z/pgGi//DD8M3wx/DFvfSDK6mjofSBv6PwikDdJGDhvfCFdeXAyfABkZP2CEGRnwoRnRoIEB9AAAAAAAAAAAAAAAAAAIGeLZMCAQH2AGHwhZGX//CHnhZ/8I2eFgGT2qj/8M8ADFuZPCot8ILdHCfaiaGn/6Z/pgGi//DD8M3wx/DFva4b/yupo6Gn/7+j8AGRF7gAAAAAAAAAAAAAAAAhni2fA58jjyxi9EOeF/+S4/YAYfCFkZf/8IeeFn/wjZ4WAZPaqP/wzwAgFIFxQBCbi3xYJQFQH8+EFujhPtRNDT/9M/0wDRf/hh+Gb4Y/hi3tcN/5XU0dDT/9/R+ADIi9wAAAAAAAAAAAAAAAAQzxbPgc+Rx5YxeiHPC//JcfsAyIvcAAAAAAAAAAAAAAAAEM8Wz4HPklb4sEohzwv/yXH7ADD4QsjL//hDzws/+EbPCwDJ7VR/FgAE+GcActxwItDWAjHSADDcIccAkvI74CHXDR+S8jzhUxGS8jvhwQQighD////9vLGS8jzgAfAB+EdukvI83g==',
            signed.message)

        # Sign with signing box
        sbox = async_core_client.crypto.get_signing_box(params=self.keypair)
        signer_sbox = Signer.SigningBox(handle=sbox.handle)
        encode_params = ParamsOfEncodeMessage(
            abi=self.events_abi, signer=signer_sbox, deploy_set=deploy_set,
            call_set=call_set)
        signed_sbox = async_core_client.abi.encode_message(
            params=encode_params)
        async_core_client.crypto.remove_signing_box(params=sbox)
        self.assertEqual(signed.message, signed_sbox.message)

        # Create run unsigned message
        address = '0:05beb555e942fa744fd96f45a9ea9d0a8248208ca12421947c06e59bc997d309'
        call_set = CallSet(
            function_name='returnValue', input={'id': '0'},
            header=FunctionHeader(**{
                'pubkey': self.keypair.public,
                'time': self.events_time,
                'expire': self.events_expire
            }))
        signer = Signer.External(public_key=self.keypair.public)
        encode_params = ParamsOfEncodeMessage(
            abi=self.events_abi, signer=signer, address=address,
            call_set=call_set)
        unsigned = async_core_client.abi.encode_message(params=encode_params)
        self.assertEqual(
            'te6ccgEBAgEAeAABpYgAC31qq9KF9Oifst6LU9U6FQSQQRlCSEMo+A3LN5MvphIFMfECP8d3ruNZAXul5xxahT91swIEkEHph08JVlwmUmQAAAXRnJcuDX1XMZBW+LBKAQBAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=',
            unsigned.message)
        self.assertEqual(
            'i4Hs3PB12QA9UBFbOIpkG3JerHHqjm4LgvF4MA7TDsY=',
            unsigned.data_to_sign)

        # Create detached signature
        sign_params = ParamsOfSign(
            unsigned=unsigned.data_to_sign, keys=self.keypair)
        signature = async_core_client.crypto.sign(params=sign_params)
        self.assertEqual(
            '5bbfb7f184f2cb5f019400b9cd497eeaa41f3d5885619e9f7d4fab8dd695f4b3a02159a1422996c1dd7d1be67898bc79c6adba6c65a18101ac5f0a2a2bb8910b',
            signature.signature)

        # Attach signature
        attach_params = ParamsOfAttachSignature(
            abi=self.events_abi, public_key=self.keypair.public,
            message=unsigned.message, signature=signature.signature)
        signed = async_core_client.abi.attach_signature(params=attach_params)
        self.assertEqual(
            'te6ccgEBAwEAvAABRYgAC31qq9KF9Oifst6LU9U6FQSQQRlCSEMo+A3LN5MvphIMAQHhrd/b+MJ5Za+AygBc5qS/dVIPnqxCsM9PvqfVxutK+lnQEKzQoRTLYO6+jfM8TF4841bdNjLQwIDWL4UVFdxIhdMfECP8d3ruNZAXul5xxahT91swIEkEHph08JVlwmUmQAAAXRnJcuDX1XMZBW+LBKACAEAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA==',
            signed.message)

        # Create initially signed message
        signer = Signer.Keys(keys=self.keypair)
        encode_params = ParamsOfEncodeMessage(
            abi=self.events_abi, signer=signer, address=address,
            call_set=call_set)
        signed = async_core_client.abi.encode_message(params=encode_params)
        self.assertEqual(
            'te6ccgEBAwEAvAABRYgAC31qq9KF9Oifst6LU9U6FQSQQRlCSEMo+A3LN5MvphIMAQHhrd/b+MJ5Za+AygBc5qS/dVIPnqxCsM9PvqfVxutK+lnQEKzQoRTLYO6+jfM8TF4841bdNjLQwIDWL4UVFdxIhdMfECP8d3ruNZAXul5xxahT91swIEkEHph08JVlwmUmQAAAXRnJcuDX1XMZBW+LBKACAEAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA==',
            signed.message)

    def test_encode_message_body(self):
        header = FunctionHeader(
            expire=self.events_expire, time=self.events_time,
            pubkey=self.keypair.public)
        call_set = CallSet(
            function_name='returnValue', header=header, input={'id': '0'})
        signer = Signer.Keys(keys=self.keypair)
        params = ParamsOfEncodeMessageBody(
            abi=self.events_abi, call_set=call_set, is_internal=False,
            signer=signer)
        encoded = async_core_client.abi.encode_message_body(params=params)

        self.assertIsNone(encoded.data_to_sign)

    def test_encode_account(self):
        # Encode account from encoded deploy message
        encoded_deploy_message = 'te6ccgECFwEAA2gAAqeIAAt9aqvShfTon7Lei1PVOhUEkEEZQkhDKPgNyzeTL6YSEZTHxAj/Hd67jWQF7peccWoU/dbMCBJBB6YdPCVZcJlJkAAAF0ZyXLg19VzGRotV8/gGAQEBwAICA88gBQMBAd4EAAPQIABB2mPiBH+O713GsgL3S844tQp+62YECSCD0w6eEqy4TKTMAib/APSkICLAAZL0oOGK7VNYMPShCQcBCvSkIPShCAAAAgEgDAoByP9/Ie1E0CDXScIBjhDT/9M/0wDRf/hh+Gb4Y/hijhj0BXABgED0DvK91wv/+GJw+GNw+GZ/+GHi0wABjh2BAgDXGCD5AQHTAAGU0/8DAZMC+ELiIPhl+RDyqJXTAAHyeuLTPwELAGqOHvhDIbkgnzAg+COBA+iogggbd0Cgud6S+GPggDTyNNjTHwH4I7zyudMfAfAB+EdukvI83gIBIBINAgEgDw4AvbqLVfP/hBbo417UTQINdJwgGOENP/0z/TANF/+GH4Zvhj+GKOGPQFcAGAQPQO8r3XC//4YnD4Y3D4Zn/4YeLe+Ebyc3H4ZtH4APhCyMv/+EPPCz/4Rs8LAMntVH/4Z4AgEgERAA5biABrW/CC3Rwn2omhp/+mf6YBov/ww/DN8Mfwxb30gyupo6H0gb+j8IpA3SRg4b3whXXlwMnwAZGT9ghBkZ8KEZ0aCBAfQAAAAAAAAAAAAAAAAACBni2TAgEB9gBh8IWRl//wh54Wf/CNnhYBk9qo//DPAAxbmTwqLfCC3Rwn2omhp/+mf6YBov/ww/DN8Mfwxb2uG/8rqaOhp/+/o/ABkRe4AAAAAAAAAAAAAAAAIZ4tnwOfI48sYvRDnhf/kuP2AGHwhZGX//CHnhZ/8I2eFgGT2qj/8M8AIBSBYTAQm4t8WCUBQB/PhBbo4T7UTQ0//TP9MA0X/4Yfhm+GP4Yt7XDf+V1NHQ0//f0fgAyIvcAAAAAAAAAAAAAAAAEM8Wz4HPkceWMXohzwv/yXH7AMiL3AAAAAAAAAAAAAAAABDPFs+Bz5JW+LBKIc8L/8lx+wAw+ELIy//4Q88LP/hGzwsAye1UfxUABPhnAHLccCLQ1gIx0gAw3CHHAJLyO+Ah1w0fkvI84VMRkvI74cEEIoIQ/////byxkvI84AHwAfhHbpLyPN4='
        message_source = MessageSource.Encoded(
            message=encoded_deploy_message, abi=self.events_abi)
        state_init_source = StateInitSource.Message(source=message_source)

        encode_params = ParamsOfEncodeAccount(state_init=state_init_source)
        encoded = async_core_client.abi.encode_account(params=encode_params)
        self.assertEqual(
            '05beb555e942fa744fd96f45a9ea9d0a8248208ca12421947c06e59bc997d309',
            encoded.id)

        # Encode account from encoding params
        deploy_set = DeploySet(tvc=self.events_tvc)
        call_set = CallSet(
            function_name='constructor',
            header=FunctionHeader(**{
                'pubkey': self.keypair.public,
                'time': self.events_time,
                'expire': self.events_expire
            }))
        signer = Signer.Keys(keys=self.keypair)
        encoding_params = ParamsOfEncodeMessage(
            abi=self.events_abi, signer=signer, deploy_set=deploy_set,
            call_set=call_set)
        message_source = MessageSource.EncodingParams(params=encoding_params)
        state_init_source = StateInitSource.Message(source=message_source)
        encode_params = ParamsOfEncodeAccount(state_init=state_init_source)
        encoded = async_core_client.abi.encode_account(params=encode_params)
        self.assertEqual(
            '05beb555e942fa744fd96f45a9ea9d0a8248208ca12421947c06e59bc997d309',
            encoded.id)

        # Test exception (external signer)
        with self.assertRaises(TonException):
            signer = Signer.External(public_key=self.keypair.public)
            encoding_params = ParamsOfEncodeMessage(
                abi=self.events_abi, signer=signer, deploy_set=deploy_set,
                call_set=call_set)
            message_source = MessageSource.EncodingParams(
                params=encoding_params)
            state_init_source = StateInitSource.Message(source=message_source)
            encode_params = ParamsOfEncodeAccount(state_init=state_init_source)
            async_core_client.abi.encode_account(params=encode_params)

        # Encode account from TVC
        state_init_source = StateInitSource.Tvc(tvc=self.events_tvc)
        encode_params = ParamsOfEncodeAccount(state_init=state_init_source)
        encoded = async_core_client.abi.encode_account(params=encode_params)
        self.assertNotEqual(
            '05beb555e942fa744fd96f45a9ea9d0a8248208ca12421947c06e59bc997d309',
            encoded.id)

        state_init_source = StateInitSource.Tvc(
            tvc=self.events_tvc, public_key=self.keypair.public)
        encode_params = ParamsOfEncodeAccount(state_init=state_init_source)
        encoded = async_core_client.abi.encode_account(params=encode_params)
        self.assertEqual(
            '05beb555e942fa744fd96f45a9ea9d0a8248208ca12421947c06e59bc997d309',
            encoded.id)


class TestTonAbiSyncCore(unittest.TestCase):
    """ Sync core is not recommended to use, so make just a couple of tests """
    def setUp(self) -> None:
        # Events contract params
        self.events_abi = Abi.from_path(
            path=os.path.join(SAMPLES_DIR, 'Events.abi.json'))
        self.keypair = KeyPair(
            public='4c7c408ff1ddebb8d6405ee979c716a14fdd6cc08124107a61d3c25597099499',
            secret='cc8929d635719612a9478b9cd17675a39cfad52d8959e8a177389b8c0b9122a7')
        with open(os.path.join(SAMPLES_DIR, 'Events.tvc'), 'rb') as fp:
            self.events_tvc = base64.b64encode(fp.read()).decode()
        self.events_time = 1599458364291
        self.events_expire = 1599458404

    def test_decode_message(self):
        message = 'te6ccgEBAwEAvAABRYgAC31qq9KF9Oifst6LU9U6FQSQQRlCSEMo+A3LN5MvphIMAQHhrd/b+MJ5Za+AygBc5qS/dVIPnqxCsM9PvqfVxutK+lnQEKzQoRTLYO6+jfM8TF4841bdNjLQwIDWL4UVFdxIhdMfECP8d3ruNZAXul5xxahT91swIEkEHph08JVlwmUmQAAAXRnJcuDX1XMZBW+LBKACAEAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=='
        params = ParamsOfDecodeMessage(abi=self.events_abi, message=message)
        decoded = sync_core_client.abi.decode_message(params=params)
        self.assertEqual('Input', decoded.body_type)
        self.assertEqual(0, int(decoded.value['id'], 16))
        self.assertEqual(self.events_expire, decoded.header.expire)
        self.assertEqual(self.events_time, decoded.header.time)
        self.assertEqual(self.keypair.public, decoded.header.pubkey)

        message = 'te6ccgEBAQEAVQAApeACvg5/pmQpY4m61HmJ0ne+zjHJu3MNG8rJxUDLbHKBu/AAAAAAAAAMJL6z6ro48sYvAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAABA'
        params = ParamsOfDecodeMessage(abi=self.events_abi, message=message)
        decoded = sync_core_client.abi.decode_message(params=params)
        self.assertEqual('Event', decoded.body_type)
        self.assertEqual(0, int(decoded.value['id'], 16))
        self.assertIsNone(decoded.header)

        message = 'te6ccgEBAQEAVQAApeACvg5/pmQpY4m61HmJ0ne+zjHJu3MNG8rJxUDLbHKBu/AAAAAAAAAMKr6z6rxK3xYJAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAABA'
        params = ParamsOfDecodeMessage(abi=self.events_abi, message=message)
        decoded = sync_core_client.abi.decode_message(params=params)
        self.assertEqual('Output', decoded.body_type)
        self.assertEqual(0, int(decoded.value['value0'], 16))
        self.assertIsNone(decoded.header)

        with self.assertRaises(TonException):
            params = ParamsOfDecodeMessage(abi=self.events_abi, message='0x0')
            sync_core_client.abi.decode_message(params=params)



import ctypes
import json


class TCStringData(ctypes.Structure):
    _fields_ = [
        ('content', ctypes.c_char_p),
        ('len', ctypes.c_int32)
    ]

    @property
    def string(self):
        if self.len:
            return self.content[:self.len].decode(errors='replace')
        return None

    @property
    def json(self):
        if self.string:
            return json.loads(self.string)
        return None

    @staticmethod
    def from_string(string: str):
        string = string.encode()
        return TCStringData(ctypes.c_char_p(string), len(string))

    def __str__(self):
        return self.string
    __repr__ = __str__


class TCSyncResponseData(ctypes.Structure):
    _fields_ = [
        ('string_data', TCStringData)
    ]

    @property
    def is_success(self):
        return not self.error

    @property
    def result(self):
        return self.string_data.json.get('result')

    @property
    def error(self):
        return self.string_data.json.get('error')

    def __str__(self):
        return self.string_data.__str__()
    __repr__ = __str__


class TCResponseType:
    Success = 0
    Error = 1
    Nop = 2
    AppRequest = 3
    AppNotify = 4
    Custom = 100


TCClientContext = ctypes.c_int32
TCResponseHandler = ctypes.CFUNCTYPE(
    ctypes.c_void_p, ctypes.c_int32, TCStringData, ctypes.c_int32,
    ctypes.c_bool)
